"""
    Main Api implementation.
"""
import datetime
from itertools import cycle
from typing import Optional, List, Union

import pytz
from pytz import timezone
import requests
import datetime
import time

from requests.auth import HTTPBasicAuth
from requests.models import Response
from requests_oauthlib.oauth2_session import OAuth2Session

from ytvs.error import ErrorCode, ErrorMessage
from ytvs.exception import PyYouTubeException
from ytvs.models import (
    AccessToken,
    UserProfile,
    ActivityListResponse,
    CaptionListResponse,
    ChannelListResponse,
    ChannelSectionResponse,
    PlaylistListResponse,
    PlaylistItemListResponse,
    VideoListResponse,
    CommentThreadListResponse,
    CommentListResponse,
    VideoCategoryListResponse,
    SearchListResponse,
    SubscriptionListResponse,
    I18nRegionListResponse,
    I18nLanguageListResponse,
    MemberListResponse,
    MembershipsLevelListResponse,
    VideoAbuseReportReasonListResponse,
)
from pyyoutube.utils.params_checker import enf_comma_separated, enf_parts


class OfficialApi(object):
    """
    Example usage:
        To create an instance of pyyoutube.Api class:

            >>> import pyyoutube
            >>> api = pyyoutube.OfficialApi(api_key="your api key")

        To get one channel info:

            >>> res = api.get_channel_info(channel_id="UC_x5XG1OV2P6uZZ5FSM9Ttw")
            >>> print(res.items[0])

        Now this api provide methods as follows:
            >>> api.get_authorization_url()
            >>> api.generate_access_token()
            >>> api.refresh_token()
            >>> api.get_channel_info()
            >>> api.get_playlist_by_id()
            >>> api.get_playlists()
            >>> api.get_playlist_item_by_id()
            >>> api.get_playlist_items()
            >>> api.get_video_by_id()
            >>> api.get_videos_by_chart()
            >>> api.get_videos_by_myrating()
            >>> api.get_comment_thread_by_id()
            >>> api.get_comment_threads()
            >>> api.get_comment_by_id()
            >>> api.get_comments()
            >>> api.get_video_categories()
            >>> api.get_subscription_by_id()
            >>> api.get_subscription_by_channel()
            >>> api.get_subscription_by_me()
            >>> api.get_activities_by_channel()
            >>> api.get_activities_by_me()
            >>> api.get_captions_by_video()
            >>> api.get_channel_sections_by_id()
            >>> api.get_channel_sections_by_channel()
            >>> api.get_i18n_regions()
            >>> api.get_i18n_languages()
            >>> api.get_video_abuse_report_reason()
            >>> api.search()
            >>> api.search_by_keywords()
            >>> api.search_by_developer()
            >>> api.search_by_mine()
            >>> api.search_by_related_video()
    """

    BASE_URL = "https://www.googleapis.com/youtube/v3/"
    AUTHORIZATION_URL = "https://accounts.google.com/o/oauth2/v2/auth"
    EXCHANGE_ACCESS_TOKEN_URL = "https://oauth2.googleapis.com/token"
    USER_INFO_URL = "https://www.googleapis.com/oauth2/v1/userinfo"

    DEFAULT_REDIRECT_URI = "https://localhost/"

    DEFAULT_SCOPE = [
        "https://www.googleapis.com/auth/youtube",
        "https://www.googleapis.com/auth/userinfo.profile",
    ]

    DEFAULT_STATE = "PyYouTube"
    DEFAULT_TIMEOUT = 10
    DEFAULT_QUOTA = 10000  # this quota reset at 00:00:00(GMT-7) every day.

    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        #api_key: Optional[str] = None,
        api_keys: Optional[List[str]] = None,
        quota_reset_time='00:00',  # 쿼터 리셋 시간 (태평양 표준시 기준 자정)
        cool_down_period=3600,  # API 키 만료 후 대기 시간 (초 단위)
        access_token: Optional[str] = None,
        timeout: Optional[int] = None,
        proxies: Optional[dict] = None,
    ) -> None:
        """
        This Api provide two method to work. Use api key or use access token.

        Args:
            client_id(str, optional):
                Your google app's ID.
            client_secret (str, optional):
                Your google app's secret.
            api_key(str, optional):
                The api key which you create from google api console.
            access_token(str, optional):
                If you not provide api key, you can do authorization to get an access token.
                If all api key and access token provided. Use access token first.
            timeout(int, optional):
                The request timeout.
            proxies(dict, optional):
                If you want use proxy, need point this param.
                param style like requests lib style.
                Refer https://2.python-requests.org//en/latest/user/advanced/#proxies

        Returns:
            YouTube Api instance.
        """
        self._client_id = client_id
        self._client_secret = client_secret
        self._access_token = access_token
        self._refresh_token = None  # This keep current user's refresh token.
        self._timeout = timeout
        self.session = requests.Session()
        self.proxies = proxies
        self.api_keys = api_keys
        self.current_key_index = 0  # 현재 사용 중인 API 키 인덱스
        self.api_key_usage = {key: {"expired":False, "reset_time": None} for key in api_keys}  # API 키 사용량 기록
        self.quota_reset_time = quota_reset_time
        self.cool_down_period = cool_down_period
        self.last_quota_reset = None
        self.next_key_switch_time = None  # 다음 API 키로 전환할 시간

        if not (
            (self._client_id and self._client_secret)
            or self.api_keys
            or self._access_token
        ):
            raise PyYouTubeException(
                ErrorMessage(
                    status_code=ErrorCode.MISSING_PARAMS,
                    message="Must specify either client key info or api key.",
                )
            )

        if self._timeout is None:
            self._timeout = self.DEFAULT_TIMEOUT

    def _update_key_usage(self):
        """태평양 표준시 자정에 모든 키의 만료 상태를 초기화합니다."""
        pst_now = datetime.datetime.now(pytz.timezone("Asia/Seoul"))
        for key, usage in self.api_key_usage.items():
            if usage["reset_time"] is None or pst_now >= usage["reset_time"]:
                usage["expired"] = False
                next_day = pst_now + datetime.timedelta(days=1)
                reset_time = datetime.datetime(year=next_day.year, month=next_day.month, day=next_day.day, hour=0, minute=0, second=0, tzinfo=pytz.timezone("America/Los_Angeles"))
                usage["reset_time"] = reset_time


    def _check_if_all_key_is_expired(self):
        is_all_expired = True
        for key, value in self.api_key_usage:
            is_expired = value['expired']
            if is_expired is False:
                all_expired = False
                break
        return is_all_expired




    def _get_next_key(self):
        """다음 사용할 API 키를 반환합니다. 모든 키가 만료되었다면 대기합니다."""
        original_index = self.current_key_index
        while True:
            current_key = self.api_keys[self.current_key_index]
            if not self.api_key_usage[current_key]["expired"]:
                return current_key
            self.current_key_index = (self.current_key_index + 1) % len(self.current_key_index)
            if self.current_key_index == original_index:  # 모든 키가 만료된 경우
                print("All API keys are expired. Waiting...")
                time.sleep(self.cool_down_period)
                self._update_key_usage()  # 만료 상태 업데이트 시도

    def _get_oauth_session(
        self,
        redirect_uri: Optional[str] = None,
        scope: Optional[List[str]] = None,
        **kwargs,
    ) -> OAuth2Session:
        """
        Build a request session for OAuth.

        Args:
            redirect_uri(str, optional)
                Determines how Google's authorization server sends a response to your app.
                If not provide will use default https://localhost/
            scope (list, optional)
                The scope you want give permission.
                If you not provide, will use default scope.
            kwargs(dict, optional)
                Some other params you want provide.

        Returns:
            OAuth2 Session
        """
        if redirect_uri is None:
            redirect_uri = self.DEFAULT_REDIRECT_URI

        if scope is None:
            scope = self.DEFAULT_SCOPE

        return OAuth2Session(
            client_id=self._client_id,
            scope=scope,
            redirect_uri=redirect_uri,
            state=self.DEFAULT_STATE,
            **kwargs,
        )

    def get_authorization_url(
        self,
        redirect_uri: Optional[str] = None,
        scope: Optional[List[str]] = None,
        **kwargs,
    ) -> (str, str):
        """
        Build authorization url to do authorize.

        Args:
            redirect_uri(str, optional)
                Determines how Google's authorization server sends a response to your app.
                If not provide will use default https://localhost/
            scope (list, optional)
                The scope you want give permission.
                If you not provide, will use default scope.
            kwargs(dict, optional)
                Some other params you want provide.

        Returns:
            The uri you can open on browser to do authorize.
        """
        oauth_session = self._get_oauth_session(
            redirect_uri=redirect_uri,
            scope=scope,
            **kwargs,
        )
        authorization_url, state = oauth_session.authorization_url(
            self.AUTHORIZATION_URL,
            access_type="offline",
            prompt="select_account",
            **kwargs,
        )

        return authorization_url, state

    def generate_access_token(
        self,
        authorization_response: str,
        redirect_uri: Optional[str] = None,
        scope: Optional[List[str]] = None,
        return_json: bool = False,
        **kwargs,
    ) -> Union[dict, AccessToken]:
        """
        Use the google auth response to get access token

        Args:
            authorization_response (str)
                The response url which google redirect.
            redirect_uri(str, optional)
                Determines how Google's authorization server sends a response to your app.
                If not provide will use default https://localhost/
            scope (list, optional)
                The scope you want give permission.
                If you not provide, will use default scope.
            return_json(bool, optional)
                The return data type. If you set True JSON data will be returned.
                False will return pyyoutube.AccessToken
            kwargs(dict, optional)
                Some other params you want provide.
        Return:
            Retrieved access token's info, pyyoutube.AccessToken instance.
        """

        oauth_session = self._get_oauth_session(
            redirect_uri=redirect_uri,
            scope=scope,
            **kwargs,
        )
        token = oauth_session.fetch_token(
            self.EXCHANGE_ACCESS_TOKEN_URL,
            client_secret=self._client_secret,
            authorization_response=authorization_response,
            proxies=self.proxies,
        )
        self._access_token = oauth_session.access_token
        self._refresh_token = oauth_session.token["refresh_token"]
        if return_json:
            return token
        else:
            return AccessToken.from_dict(token)

    def refresh_token(
        self, refresh_token: Optional[str] = None, return_json: bool = False
    ) -> Union[dict, AccessToken]:
        """
        Refresh token by api return refresh token.

        Args:
            refresh_token (str)
                The refresh token which the api returns.
            return_json (bool, optional):
                If True JSON data will be returned, instead of pyyoutube.AccessToken
        Return:
            Retrieved new access token's info,  pyyoutube.AccessToken instance.
        """

        refresh_token = refresh_token if refresh_token else self._refresh_token

        if refresh_token is None:
            raise PyYouTubeException(
                ErrorMessage(
                    status_code=ErrorCode.MISSING_PARAMS,
                    message=f"Must provide the refresh token or api has been authorized.",
                )
            )

        oauth_session = OAuth2Session(client_id=self._client_id)
        auth = HTTPBasicAuth(self._client_id, self._client_secret)
        new_token = oauth_session.refresh_token(
            self.EXCHANGE_ACCESS_TOKEN_URL,
            refresh_token=refresh_token,
            auth=auth,
        )
        self._access_token = oauth_session.access_token
        if return_json:
            return new_token
        else:
            return AccessToken.from_dict(new_token)

    @staticmethod
    def _parse_response(response: Response) -> dict:
        """
        Parse response data and check whether errors exists.

        Args:
            response (Response)
                The response which the request return.
        Return:
             response's data
        """
        data = response.json()
        if "error" in data:
            raise PyYouTubeException(response)
        return data

    @staticmethod
    def _parse_data(data: Optional[dict]) -> Union[dict, list]:
        """
        Parse resp data.

        Args:
            data (dict)
                The response data by response.json()
        Return:
             response's items
        """
        items = data["items"]
        return items

    def _request(
        self, resource, method=None, args=None, post_args=None, enforce_auth=True
    ) -> Response:
        """
        Main request sender.

        Args:
            resource(str)
                Resource field is which type data you want to retrieve.
                Such as channels，videos and so on.
            method(str, optional)
                The method this request to send request.
                Default is 'GET'
            args(dict, optional)
                The url params for this request.
            post_args(dict, optional)
                The Post params for this request.
            enforce_auth(bool, optional)
                Whether use google credentials
        Returns:
            response
        """
        if method is None:
            method = "GET"

        if args is None:
            args = dict()

        if post_args is not None:
            method = "POST"

        key = None
        access_token = None

        self._update_key_usage()  # 필요한 경우 쿼터 초기화
        api_key = self.api_keys[self.current_key_index]

        if api_key is not None:
            key = "key"
            access_token = api_key
        if self._access_token is not None:
            key = "access_token"
            access_token = self._access_token
        if access_token is None and enforce_auth:
            raise PyYouTubeException(
                ErrorMessage(
                    status_code=ErrorCode.MISSING_PARAMS,
                    message="You must provide your credentials.",
                )
            )

        if enforce_auth:
            if method == "POST" and key not in post_args:
                post_args[key] = access_token
            elif method == "GET" and key not in args:
                args[key] = access_token

        try:
            response = self.session.request(
                method=method,
                url=self.BASE_URL + resource,
                timeout=self._timeout,
                params=args,
                data=post_args,
                proxies=self.proxies,
            )
        except requests.HTTPError as e:
            if e.response.status_code == 403:  # 쿼터 초과 에러 처리
                # 현재 API 키의 쿼터가 초과되었다고 표시하고 다음 키로 전환하는 로직 추가
                self.api_key_usage[api_key] = {
                    'expired': True,
                    'reset_time': datetime.datetime.now()
                }
                is_all_key_expired = self._check_if_all_key_is_expired()
                if is_all_key_expired is True:
                    raise Exception("All key is expired ")
                else:
                    return self._request(resource, method, args, post_args, enforce_auth)
            else:
                raise PyYouTubeException(
                    ErrorMessage(status_code=ErrorCode.HTTP_ERROR, message=e.args[0])
                )
        else:
            return response

    def get_profile(
        self, access_token: Optional[str] = None, return_json: Optional[bool] = False
    ) -> Union[dict, UserProfile]:
        """
        Get token user info.

        Args:
            access_token(str, optional)
                user access token. If not provide, use api instance access token
            return_json(bool, optional)
                The return data type. If you set True JSON data will be returned.
                False will return pyyoutube.UserProfile

        Returns:
            The data for you given access token's user info.
        """
        if access_token is None:
            access_token = self._access_token
        if access_token is None:
            raise PyYouTubeException(
                ErrorMessage(
                    status_code=ErrorCode.MISSING_PARAMS,
                    message=f"Must provide the access token or api has been authorized.",
                )
            )
        try:
            response = self.session.get(
                self.USER_INFO_URL,
                params={"access_token": access_token},
                timeout=self._timeout,
                proxies=self.proxies,
            )
        except requests.HTTPError as e:
            raise PyYouTubeException(
                ErrorMessage(status_code=ErrorCode.HTTP_ERROR, message=e.args[0])
            )
        data = self._parse_response(response)
        if return_json:
            return data
        else:
            return UserProfile.from_dict(data)

    def paged_by_page_token(
        self,
        resource: str,
        args: dict,
        count: Optional[int] = None,
    ):
        """
        Response paged by response's page token. If not provide response token

        Args:
            resource (str):
                The resource string need to retrieve data.
            args (dict)
                The args for api.
            count (int, optional):
                The count for result items you want to get.
                If provide this with None, will retrieve all items.
                Note:
                    The all items maybe too much. Notice your app's cost.
        Returns:
            Data api origin response.
        """
        res_data: Optional[dict] = None
        current_items: List[dict] = []
        page_token: Optional[str] = None
        now_items_count: int = 0

        while True:
            if page_token is not None:
                args["pageToken"] = page_token

            resp = self._request(resource=resource, method="GET", args=args)
            data = self._parse_response(resp)  # origin response
            # set page token
            page_token = data.get("nextPageToken")
            prev_page_token = data.get("prevPageToken")

            # parse results.
            items = self._parse_data(data)
            current_items.extend(items)
            now_items_count += len(items)
            if res_data is None:
                res_data = data
            # first check the count if satisfies.
            if count is not None:
                if now_items_count >= count:
                    current_items = current_items[:count]
                    break
            # if have no page token, mean no more data.
            if page_token is None:
                break
        res_data["items"] = current_items

        # use last request page token
        res_data["nextPageToken"] = page_token
        res_data["prevPageToken"] = prev_page_token
        return res_data

    def get_activities_by_channel(
        self,
        *,
        channel_id: str,
        parts: Optional[Union[str, list, tuple, set]] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        region_code: Optional[str] = None,
        count: Optional[int] = 20,
        limit: int = 20,
        page_token: Optional[str] = None,
        return_json: bool = False,
    ):
        """
        Retrieve given channel's activities data.

        Args:
            channel_id (str):
                The id for channel which you want to get activities data.
            parts ((str,list,tuple,set) optional):
                The resource parts for activities you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            before (str, optional):
                Set this will only return the activities occurred before this timestamp.
                This need specified in ISO 8601 (YYYY-MM-DDThh:mm:ss.sZ) format.
            after (str, optional):
                Set this will only return the activities occurred after this timestamp.
                This need specified in ISO 8601 (YYYY-MM-DDThh:mm:ss.sZ) format.
            region_code (str, optional):
                Set this will only return the activities for the specified country.
                This need specified with an ISO 3166-1 alpha-2 country code.
            count (int, optional):
                The count will retrieve activities data.
                Default is 20.
                If provide this with None, will retrieve all activities.
            limit (int, optional):
                The maximum number of items each request retrieve.
                For activities, this should not be more than 50.
                Default is 20.
            page_token (str, optional):
                The token of the page of activities result to retrieve.
                You can use this retrieve point result page directly.
                And you should know about the page result set for YouTube.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.ActivityListResponse instance.

        Returns:
            ActivityListResponse or original data.
        """

        if count is None:
            limit = 50  # for activities the max limit for per request is 50
        else:
            limit = min(count, limit)

        args = {
            "channelId": channel_id,
            "part": enf_parts(resource="activities", value=parts),
            "maxResults": limit,
        }

        if before:
            args["publishedBefore"] = before
        if after:
            args["publishedAfter"] = after
        if region_code:
            args["regionCode"] = region_code

        if page_token is not None:
            args["pageToken"] = page_token

        res_data = self.paged_by_page_token(
            resource="activities", args=args, count=count
        )

        if return_json:
            return res_data
        else:
            return ActivityListResponse.from_dict(res_data)

    def get_activities_by_me(
        self,
        *,
        parts: Optional[Union[str, list, tuple, set]] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        region_code: Optional[str] = None,
        count: Optional[int] = 20,
        limit: int = 20,
        page_token: Optional[str] = None,
        return_json: bool = False,
    ):
        """
        Retrieve authorized user's activities.

        Note:
            This need you do authorize first.

        Args:
            parts ((str,list,tuple,set) optional):
                The resource parts for activities you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            before (str, optional):
                Set this will only return the activities occurred before this timestamp.
                This need specified in ISO 8601 (YYYY-MM-DDThh:mm:ss.sZ) format.
            after (str, optional):
                Set this will only return the activities occurred after this timestamp.
                This need specified in ISO 8601 (YYYY-MM-DDThh:mm:ss.sZ) format.
            region_code (str, optional):
                Set this will only return the activities for the specified country.
                This need specified with an ISO 3166-1 alpha-2 country code.
            count (int, optional):
                The count will retrieve activities data.
                Default is 20.
                If provide this with None, will retrieve all activities.
            limit (int, optional):
                The maximum number of items each request retrieve.
                For activities, this should not be more than 50.
                Default is 20.
            page_token (str, optional):
                The token of the page of activities result to retrieve.
                You can use this retrieve point result page directly.
                And you should know about the page result set for YouTube.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.ActivityListResponse instance.

        Returns:
            ActivityListResponse or original data.
        """

        if count is None:
            limit = 50  # for activities the max limit for per request is 50
        else:
            limit = min(count, limit)

        args = {
            "mine": True,
            "part": enf_parts(resource="activities", value=parts),
            "maxResults": limit,
        }

        if before:
            args["publishedBefore"] = before
        if after:
            args["publishedAfter"] = after
        if region_code:
            args["regionCode"] = region_code

        if page_token is not None:
            args["pageToken"] = page_token

        res_data = self.paged_by_page_token(
            resource="activities", args=args, count=count
        )

        if return_json:
            return res_data
        else:
            return ActivityListResponse.from_dict(res_data)

    def get_captions_by_video(
        self,
        *,
        video_id: str,
        parts: Optional[Union[str, list, tuple, set]] = None,
        caption_id: Optional[Union[str, list, tuple, set]] = None,
        return_json: bool = False,
    ):
        """
        Retrieve authorized user's video's caption data.

        Note:
            This need you do authorize first.

        Args:
            video_id (str):
                The id for video which you want to get caption.
            parts ((str,list,tuple,set) optional):
                The resource parts for caption you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            caption_id ((str,list,tuple,set)):
                The id for caption that you want to get data.
                You can pass this with single id str,comma-separated id str, or list, tuple, set of id str.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.CaptionListResponse instance.
        Returns:
            CaptionListResponse or original data.
        """

        args = {
            "videoId": video_id,
            "part": enf_parts("captions", parts),
        }

        if caption_id is not None:
            args["id"] = enf_comma_separated("caption_id", caption_id)

        resp = self._request(resource="captions", method="GET", args=args)
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return CaptionListResponse.from_dict(data)

    def get_channel_info(
        self,
        # *,
        channel_id: Optional[Union[str, list, tuple, set]] = None,
        for_username: Optional[str] = None,
        mine: Optional[bool] = None,
        parts: Optional[Union[str, list, tuple, set]] = None,
        hl: str = "en_US",
        return_json: Optional[bool] = False
    ):
        """
        Retrieve channel data from YouTube Data API.

        Note:
            1. Don't know why, but now you couldn't get channel list by given an guide category.
               You can only get list by parameters mine,forUsername,id.
               Refer: https://developers.google.com/youtube/v3/guides/implementation/channels
            2. The origin maxResult param not work for these filter method.

        Args:
            channel_id ((str,list,tuple,set), optional):
                The id or comma-separated id string for youtube channel which you want to get.
                You can also pass this with an id list, tuple, set.
            for_username (str, optional):
                The name for YouTube username which you want to get.
                Note: This name may the old youtube version's channel's user's username, Not the the channel name.
                Refer: https://developers.google.com/youtube/v3/guides/working_with_channel_ids
            mine (bool, optional):
                If you have give the authorization. Will return your channels.
                Must provide the access token.
            parts (str, optional):
                Comma-separated list of one or more channel resource properties.
                If not provided. will use default public properties.
            hl (str, optional):
                If provide this. Will return channel's language localized info.
                This value need https://developers.google.com/youtube/v3/docs/i18nLanguages.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.ChannelListResponse instance.
        Returns:
            ChannelListResponse instance or original data.
        """

        args = {
            "part": enf_parts(resource="channels", value=parts),
            "hl": hl,
        }
        if for_username is not None:
            args["forUsername"] = for_username
        elif channel_id is not None:
            args["id"] = enf_comma_separated("channel_id", channel_id)
        elif mine is not None:
            args["mine"] = mine
        else:
            raise PyYouTubeException(
                ErrorMessage(
                    status_code=ErrorCode.MISSING_PARAMS,
                    message=f"Specify at least one of channel_id,channel_name or mine",
                )
            )

        resp = self._request(resource="channels", method="GET", args=args)

        data = self._parse_response(resp)
        if return_json:
            return data
        else:
            return ChannelListResponse.from_dict(data)

    def get_channel_sections_by_id(
        self,
        *,
        section_id: Union[str, list, tuple, set],
        parts: Optional[Union[str, list, tuple, set]] = None,
        return_json: Optional[bool] = False,
    ) -> Union[ChannelSectionResponse, dict]:
        """
        Retrieve channel section info by his ids(s).

        Args:
            section_id:
                The id(s) for channel sections.
                You can pass this with single id str, comma-separated id str, or a list,tuple,set of ids.
            parts:
                The resource parts for channel section you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            return_json:
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.ChannelSectionResponse instance.
        Returns:
            ChannelSectionResponse or original data.
        """

        args = {
            "id": enf_comma_separated(field="section_id", value=section_id),
            "part": enf_parts(resource="channelSections", value=parts),
        }

        resp = self._request(resource="channelSections", args=args)
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return ChannelSectionResponse.from_dict(data)

    def get_channel_sections_by_channel(
        self,
        *,
        channel_id: Optional[str] = None,
        mine: bool = False,
        parts: Optional[Union[str, list, tuple, set]] = None,
        return_json: Optional[bool] = False,
    ) -> Union[ChannelSectionResponse, dict]:
        """
        Retrieve channel sections by channel id.

        Args:
            channel_id:
                The id for channel which you want to get channel sections.
            mine:
                If you want to get your channel's sections, set this with True.
                And this need your authorization.
            parts:
                The resource parts for channel section you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            return_json:
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.ChannelSectionResponse instance.
        Returns:
            ChannelSectionResponse or original data.
        """

        args = {
            "part": enf_parts(resource="channelSections", value=parts),
        }

        if mine:
            args["mine"] = mine
        else:
            args["channelId"] = channel_id

        resp = self._request(resource="channelSections", args=args)
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return ChannelSectionResponse.from_dict(data)

    def get_comment_by_id(
        self,
        *,
        comment_id: Union[str, list, tuple, set],
        parts: Optional[Union[str, list, tuple, set]] = None,
        text_format: Optional[str] = "html",
        return_json: Optional[bool] = False,
    ):
        """
        Retrieve comment info by given comment id str.

        Args:
            comment_id (str, optional):
                The id for comment that you want to retrieve data.
                You can pass this with single id str, comma-separated id str, or a list,tuple,set of ids.
            parts ((str,list,tuple,set), optional):
                The resource parts for you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            text_format (str, optional):
                Comments left by users format style.
                Acceptable values are: html, plainText.
                Default is html.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.CommentListResponse instance.

        Returns:
            CommentListResponse or original data
        """

        args = {
            "id": enf_comma_separated(field="comment_id", value=comment_id),
            "part": enf_parts(resource="comments", value=parts),
            "textFormat": text_format,
        }

        resp = self._request(resource="comments", method="GET", args=args)
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return CommentListResponse.from_dict(data)

    def get_comments(
        self,
        *,
        parent_id: str,
        parts: Optional[Union[str, list, tuple, set]] = None,
        text_format: Optional[str] = "html",
        count: Optional[int] = 20,
        limit: Optional[int] = 20,
        page_token: Optional[str] = None,
        return_json: Optional[bool] = False,
    ):
        """
        Retrieve comments info by given parent id.
        Note: YouTube currently supports replies only for top-level comments.
        However, replies to replies may be supported in the future.

        Args:
            parent_id (str):
                Provide the ID of the comment for which replies should be retrieved.
            parts ((str,list,tuple,set), optional):
                The resource parts for you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            text_format (str, optional):
                Comments left by users format style.
                Acceptable values are: html, plainText.
                Default is html.
            count (int, optional):
                The count will retrieve videos data.
                Default is 20.
                If provide this with None, will retrieve all comments.
            limit (int, optional):
                The maximum number of items each request retrieve.
                For comments, this should not be more than 100.
                Default is 20.
            page_token(str, optional):
                The token of the page of comments result to retrieve.
                You can use this retrieve point result page directly.
                And you should know about the the result set for YouTube.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.CommentListResponse instance.
        Returns:
            CommentListResponse or original data
        """

        if count is None:
            limit = 100  # for comments the max limit for per request is 100
        else:
            limit = min(count, limit)

        args = {
            "parentId": parent_id,
            "part": enf_parts(resource="comments", value=parts),
            "textFormat": text_format,
            "maxResults": limit,
        }

        if page_token is not None:
            args["pageToken"] = page_token

        res_data = self.paged_by_page_token(resource="comments", args=args, count=count)
        if return_json:
            return res_data
        else:
            return CommentListResponse.from_dict(res_data)

    def get_comment_thread_by_id(
        self,
        *,
        comment_thread_id: Union[str, list, tuple, set],
        parts: Optional[Union[str, list, tuple, set]] = None,
        text_format: Optional[str] = "html",
        return_json: Optional[bool] = False,
    ):
        """
        Retrieve the comment thread info by given id.

        Args:
            comment_thread_id ((str,list,tuple,set)):
                The id for comment thread that you want to retrieve data.
                You can pass this with single id str, comma-separated id str, or a list,tuple,set of ids.
            parts ((str,list,tuple,set), optional):
                The resource parts for you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            text_format (str, optional):
                Comments left by users format style.
                Acceptable values are: html, plainText.
                Default is html.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.CommentThreadListResponse instance.
        Returns:
            CommentThreadListResponse or original data
        """

        args = {
            "id": enf_comma_separated("comment_thread_id", comment_thread_id),
            "part": enf_parts(resource="commentThreads", value=parts),
            "textFormat": text_format,
        }

        resp = self._request(resource="commentThreads", method="GET", args=args)
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return CommentThreadListResponse.from_dict(data)

    def get_comment_threads(
        self,
        *,
        all_to_channel_id: Optional[str] = None,
        channel_id: Optional[str] = None,
        video_id: Optional[str] = None,
        parts: Optional[Union[str, list, tuple, set]] = None,
        moderation_status: Optional[str] = None,
        order: Optional[str] = None,
        search_terms: Optional[str] = None,
        text_format: Optional[str] = "html",
        count: Optional[int] = 20,
        limit: Optional[int] = 20,
        page_token: Optional[str] = None,
        return_json: Optional[bool] = False,
    ):
        """
        Retrieve the comment threads info by given filter condition.

        Args:
            all_to_channel_id (str, optional):
                If you provide this with a channel id, will return all comment threads associated with the channel.
                The response can include comments about the channel or about the channel's videos.
            channel_id (str, optional):
                If you provide this with a channel id, will return the comment threads associated with the channel.
                But the response not include comments about the channel's videos.
            video_id  (str, optional):
                If you provide this with a video id, will return the comment threads associated with the video.
            parts ((str,list,tuple,set), optional)
                The resource parts for you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            moderation_status (str, optional):
                This parameter must used with authorization.
                If you provide this. the response will return comment threads match this filter only.
                Acceptable values are:
                    - heldForReview: Retrieve comment threads that are awaiting review by a moderator.
                    - likelySpam: Retrieve comment threads classified as likely to be spam.
                    - published: Retrieve threads of published comments. this is default for all.
                See more: https://developers.google.com/youtube/v3/docs/commentThreads/list#parameters
            order (str, optional):
                Order parameter specifies the order in which the API response should list comment threads.
                Acceptable values are:
                    - time: Comment threads are ordered by time. This is the default behavior.
                    - relevance: Comment threads are ordered by relevance.
            search_terms (str, optional):
                The searchTerms parameter instructs the API to limit the API response to only contain comments
                that contain the specified search terms.
            text_format (str, optional):
                Comments left by users format style.
                Acceptable values are: html, plainText.
                Default is html.
            count (int, optional):
                The count will retrieve comment threads data.
                Default is 20.
                If provide this with None, will retrieve all comment threads.
            limit (int, optional):
                The maximum number of items each request retrieve.
                For comment threads, this should not be more than 100.
                Default is 20.
            page_token(str, optional):
                The token of the page of commentThreads result to retrieve.
                You can use this retrieve point result page directly.
                And you should know about the the result set for YouTube.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.CommentThreadListResponse instance.

        Returns:
            CommentThreadListResponse or original data
        """

        if count is None:
            limit = 100  # for commentThreads the max limit for per request is 100
        else:
            limit = min(count, limit)

        args = {
            "part": enf_parts(resource="commentThreads", value=parts),
            "maxResults": limit,
            "textFormat": text_format,
        }

        if all_to_channel_id:
            args["allThreadsRelatedToChannelId"] = (all_to_channel_id,)
        elif channel_id:
            args["channelId"] = channel_id
        elif video_id:
            args["videoId"] = video_id
        else:
            raise PyYouTubeException(
                ErrorMessage(
                    status_code=ErrorCode.MISSING_PARAMS,
                    message=f"Specify at least one of all_to_channel_id, channel_id or video_id",
                )
            )

        if moderation_status:
            args["moderationStatus"] = moderation_status
        if order:
            args["order"] = order
        if search_terms:
            args["searchTerms"] = search_terms

        if page_token is not None:
            args["pageToken"] = page_token

        res_data = self.paged_by_page_token(
            resource="commentThreads", args=args, count=count
        )
        if return_json:
            return res_data
        else:
            return CommentThreadListResponse.from_dict(res_data)

    def get_i18n_languages(
        self,
        *,
        parts: Optional[Union[str, list, tuple, set]] = None,
        hl: Optional[str] = "en_US",
        return_json: Optional[bool] = False,
    ) -> Union[I18nLanguageListResponse, dict]:
        """
        Returns a list of application languages that the YouTube website supports.

        Args:
            parts:
                The resource parts for i18n language you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            hl:
                If provide this. Will return i18n language's language localized info.
                This value need https://developers.google.com/youtube/v3/docs/i18nLanguages.
            return_json:
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.I18nLanguageListResponse instance.

        Returns:
            I18nLanguageListResponse or original data.
        """

        args = {"hl": hl, "part": enf_parts(resource="i18nLanguages", value=parts)}

        resp = self._request(resource="i18nLanguages", args=args)
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return I18nLanguageListResponse.from_dict(data)

    def get_i18n_regions(
        self,
        *,
        parts: Optional[Union[str, list, tuple, set]] = None,
        hl: Optional[str] = "en_US",
        return_json: Optional[bool] = False,
    ) -> Union[I18nRegionListResponse, dict]:
        """
        Retrieve all available regions.

        Args:
            parts:
                The resource parts for i18n region you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            hl:
                If provide this. Will return i18n region's language localized info.
                This value need https://developers.google.com/youtube/v3/docs/i18nLanguages.
            return_json:
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.I18nRegionListResponse instance.
        Returns:
            I18nRegionListResponse or origin data
        """

        args = {"hl": hl, "part": enf_parts(resource="i18nRegions", value=parts)}

        resp = self._request(resource="i18nRegions", args=args)
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return I18nRegionListResponse.from_dict(data)

    def get_members(
        self,
        *,
        parts: Optional[Union[str, list, tuple, set]] = None,
        mode: Optional[str] = "all_current",
        count: Optional[int] = 5,
        limit: Optional[int] = 5,
        page_token: Optional[str] = None,
        has_access_to_level: Optional[str] = None,
        filter_by_member_channel_id: Optional[Union[str, list, tuple, set]] = None,
        return_json: Optional[bool] = False,
    ) -> Union[MemberListResponse, dict]:
        """
        Retrieve a list of members for a channel.

        Args:
            parts ((str,list,tuple,set) optional):
                The resource parts for member you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            mode:
                The mode parameter indicates which members will be included in the API response.
                Set the parameter value to one of the following values:
                    - all_current (default): List current members, from newest to oldest. When this value is used,
                        the end of the list is reached when the API response does not contain a nextPageToken.
                    - updates : List only members that joined or upgraded since the previous API call.
                        Note: The first call starts a new stream of updates but does not actually return any members.
                        To start retrieving the membership updates, you need to poll the endpoint using the
                        nextPageToken at your desired frequency.
                        Note that when this value is used, the API response always contains a nextPageToken.
            count (int, optional):
                The count will retrieve videos data.
                Default is 5.
            limit (int, optional):
                The maximum number of items each request retrieve.
                For members, this should not be more than 1000.
                Default is 5.
            page_token (str, optional):
                The token of the page of search result to retrieve.
                You can use this retrieve point result page directly.
                And you should know about the the result set for YouTube.
            has_access_to_level (str, optional):
                The hasAccessToLevel parameter value is a level ID that specifies the minimum level
                that members in the result set should have.
            filter_by_member_channel_id ((str,list,tuple,set) optional):
                A list of channel IDs that can be used to check the membership status of specific users.
                A maximum of 100 channels can be specified per call.
            return_json (bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.MemberListResponse instance.
        Returns:
            MemberListResponse or original data
        """

        if count is None:
            limit = 1000
        else:
            limit = min(count, limit)

        args = {
            "part": enf_parts(resource="members", value=parts),
            "maxResults": limit,
        }

        if mode:
            args["mode"] = mode

        if page_token is not None:
            args["pageToken"] = page_token

        if has_access_to_level:
            args["hasAccessToLevel"] = has_access_to_level

        if filter_by_member_channel_id:
            args["filterByMemberChannelId"] = enf_parts(
                resource="filterByMemberChannelId",
                value=filter_by_member_channel_id,
                check=False,
            )

        res_data = self.paged_by_page_token(
            resource="members",
            args=args,
            count=count,
        )
        if return_json:
            return res_data
        else:
            return MemberListResponse.from_dict(res_data)

    def get_membership_levels(
        self,
        *,
        parts: Optional[Union[str, list, tuple, set]] = None,
        return_json: Optional[bool] = False,
    ) -> Union[MembershipsLevelListResponse, dict]:
        """
        Retrieve membership levels for a channel

        Notes:
            This requires your authorization.

        Args:
            parts ((str,list,tuple,set) optional):
                The resource parts for membership level you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            return_json (bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.MembershipsLevelListResponse instance.

        Returns:
            MembershipsLevelListResponse or original data
        """

        args = {
            "part": enf_parts(resource="membershipsLevels", value=parts),
        }

        resp = self._request(resource="membershipsLevels", args=args)
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return MembershipsLevelListResponse.from_dict(data)

    def get_playlist_item_by_id(
        self,
        *,
        playlist_item_id: Union[str, list, tuple, set],
        parts: Optional[Union[str, list, tuple, set]] = None,
        return_json: Optional[bool] = False,
    ):
        """
        Retrieve playlist Items info by your given id

        Args:
            playlist_item_id ((str,list,tuple,set)):
                The id for playlist item that you want to retrieve info.
                You can pass this with single id str, comma-separated id str.
                Or a list,tuple,set of ids.
            parts ((str,list,tuple,set) optional):
                The resource parts for you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.PlayListItemApiResponse instance.
        Returns:
            PlaylistItemListResponse or original data
        """

        args = {
            "id": enf_comma_separated("playlist_item_id", playlist_item_id),
            "part": enf_parts(resource="playlistItems", value=parts),
        }

        resp = self._request(resource="playlistItems", method="GET", args=args)
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return PlaylistItemListResponse.from_dict(data)

    def get_playlist_items(
        self,
        *,
        playlist_id: str,
        parts: Optional[Union[str, list, tuple, set]] = None,
        video_id: Optional[str] = None,
        count: Optional[int] = 5,
        limit: Optional[int] = 5,
        page_token: Optional[str] = None,
        return_json: Optional[bool] = False,
    ):
        """
        Retrieve playlist Items info by your given playlist id

        Args:
            playlist_id (str):
                The id for playlist that you want to retrieve items data.
            parts ((str,list,tuple,set) optional):
                The resource parts for you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            video_id (str, Optional):
                Specifies that the request should return only the playlist items that contain the specified video.
            count (int, optional):
                The count will retrieve playlist items data.
                Default is 5.
                If provide this with None, will retrieve all playlist items.
            limit (int, optional):
                The maximum number of items each request retrieve.
                For playlistItem, this should not be more than 50.
                Default is 5
            page_token(str, optional):
                The token of the page of playlist items result to retrieve.
                You can use this retrieve point result page directly.
                And you should know about the the result set for YouTube.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.PlayListItemApiResponse instance.
        Returns:
            PlaylistItemListResponse or original data
        """

        if count is None:
            limit = 50  # for playlistItems the max limit for per request is 50
        else:
            limit = min(count, limit)

        args = {
            "playlistId": playlist_id,
            "part": enf_parts(resource="playlistItems", value=parts),
            "maxResults": limit,
        }
        if video_id is not None:
            args["videoId"] = video_id

        if page_token is not None:
            args["pageToken"] = page_token

        res_data = self.paged_by_page_token(
            resource="playlistItems", args=args, count=count
        )
        if return_json:
            return res_data
        else:
            return PlaylistItemListResponse.from_dict(res_data)

    def get_playlist_by_id(
        self,
        *,
        playlist_id: Union[str, list, tuple, set],
        parts: Optional[Union[str, list, tuple, set]] = None,
        hl: Optional[str] = "en_US",
        return_json: Optional[bool] = False,
    ):
        """
        Retrieve playlist data by given playlist id.

        Args:
            playlist_id ((str,list,tuple,set)):
                The id for playlist that you want to retrieve data.
                You can pass this with single id str,comma-separated id str, or list, tuple, set of id str.
            parts (str, optional):
                Comma-separated list of one or more playlist resource properties.
                You can also pass this with list, tuple, set of part str.
                If not provided. will use default public properties.
            hl (str, optional):
                If provide this. Will return playlist's language localized info.
                This value need https://developers.google.com/youtube/v3/docs/i18nLanguages.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.PlaylistListResponse instance
        Returns:
            PlaylistListResponse or original data
        """
        args = {
            "id": enf_comma_separated("playlist_id", playlist_id),
            "part": enf_parts(resource="playlists", value=parts),
            "hl": hl,
        }

        resp = self._request(resource="playlists", method="GET", args=args)
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return PlaylistListResponse.from_dict(data)

    def get_playlists(
        self,
        *,
        channel_id: Optional[str] = None,
        mine: Optional[bool] = None,
        parts: Optional[Union[str, list, tuple, set]] = None,
        count: Optional[int] = 5,
        limit: Optional[int] = 5,
        hl: Optional[str] = "en_US",
        page_token: Optional[str] = None,
        return_json: Optional[bool] = False,
    ):
        """
        Retrieve channel playlists info from youtube data api.

        Args:
            channel_id (str, optional):
                If provide channel id, this will return pointed channel's playlist info.
            mine (bool, optional):
                If you have given the authorization. Will return your playlists.
                Must provide the access token.
            parts (str, optional):
                Comma-separated list of one or more playlist resource properties.
                You can also pass this with list, tuple, set of part str.
                If not provided. will use default public properties.
            count (int, optional):
                The count will retrieve playlist data.
                Default is 5.
                If provide this with None, will retrieve all playlists.
            limit (int, optional):
                The maximum number of items each request to retrieve.
                For playlist, this should not be more than 50.
                Default is 5
            hl (str, optional):
                If provide this. Will return playlist's language localized info.
                This value need https://developers.google.com/youtube/v3/docs/i18nLanguages.
            page_token(str, optional):
                The token of the page of playlists result to retrieve.
                You can use this retrieve point result page directly.
                And you should know about the the result set for YouTube.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.PlaylistListResponse instance.
        Returns:
            PlaylistListResponse or original data
        """

        if count is None:
            limit = 50  # for playlists the max limit for per request is 50
        else:
            limit = min(count, limit)

        args = {
            "part": enf_parts(resource="playlists", value=parts),
            "hl": hl,
            "maxResults": limit,
        }

        if channel_id is not None:
            args["channelId"] = channel_id
        elif mine is not None:
            args["mine"] = mine
        else:
            raise PyYouTubeException(
                ErrorMessage(
                    status_code=ErrorCode.MISSING_PARAMS,
                    message=f"Specify at least one of channel_id,playlist_id or mine",
                )
            )

        if page_token is not None:
            args["pageToken"] = page_token

        res_data = self.paged_by_page_token(
            resource="playlists", args=args, count=count
        )
        if return_json:
            return res_data
        else:
            return PlaylistListResponse.from_dict(res_data)

    def search(
        self,
        *,
        parts: Optional[Union[str, list, tuple, set]] = None,
        for_developer: Optional[bool] = None,
        for_mine: Optional[bool] = None,
        related_to_video_id: Optional[str] = None,
        channel_id: Optional[str] = None,
        channel_type: Optional[str] = None,
        event_type: Optional[str] = None,
        location: Optional[str] = None,
        location_radius: Optional[str] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        order: Optional[str] = None,
        published_after: Optional[str] = None,
        published_before: Optional[str] = None,
        q: Optional[str] = None,
        region_code: Optional[str] = None,
        relevance_language: Optional[str] = None,
        safe_search: Optional[str] = None,
        topic_id: Optional[str] = None,
        search_type: Optional[Union[str, list, tuple, set]] = None,
        video_caption: Optional[str] = None,
        video_category_id: Optional[str] = None,
        video_definition: Optional[str] = None,
        video_dimension: Optional[str] = None,
        video_duration: Optional[str] = None,
        video_embeddable: Optional[str] = None,
        video_license: Optional[str] = None,
        video_paid_product_placement: Optional[str] = None,
        video_syndicated: Optional[str] = None,
        video_type: Optional[str] = None,
        page_token: Optional[str] = None,
        return_json: Optional[bool] = False,
    ) -> Union[SearchListResponse, dict]:
        """
        Main search api implementation.
        You can find all parameters description at https://developers.google.com/youtube/v3/docs/search/list#parameters

        Returns:
            SearchListResponse or original data
        """
        parts = enf_parts(resource="search", value=parts)
        if search_type is None:
            search_type = "video,channel,playlist"
        else:
            search_type = enf_comma_separated(field="search_type", value=search_type)

        args = {
            "part": parts,
            "maxResults": min(limit, count),
        }
        if for_developer:
            args["forDeveloper"] = for_developer
        if for_mine:
            args["forMine"] = for_mine
        if related_to_video_id:
            args["relatedToVideoId"] = related_to_video_id
        if channel_id:
            args["channelId"] = channel_id
        if channel_type:
            args["channelType"] = channel_type
        if event_type:
            args["eventType"] = event_type
        if location:
            args["location"] = location
        if location_radius:
            args["locationRadius"] = location_radius
        if order:
            args["order"] = order
        if published_after:
            args["publishedAfter"] = published_after
        if published_before:
            args["publishedBefore"] = published_before
        if q:
            args["q"] = q
        if region_code:
            args["regionCode"] = region_code
        if relevance_language:
            args["relevanceLanguage"] = relevance_language
        if safe_search:
            args["safeSearch"] = safe_search
        if topic_id:
            args["topicId"] = topic_id
        if search_type:
            args["type"] = search_type
        if video_caption:
            args["videoCaption"] = video_caption
        if video_category_id:
            args["videoCategoryId"] = video_category_id
        if video_definition:
            args["videoDefinition"] = video_definition
        if video_dimension:
            args["videoDimension"] = video_dimension
        if video_duration:
            args["videoDuration"] = video_duration
        if video_embeddable:
            args["videoEmbeddable"] = video_embeddable
        if video_license:
            args["videoLicense"] = video_license
        if video_paid_product_placement:
            args["videoPaidProductPlacement"] = video_paid_product_placement
        if video_syndicated:
            args["videoSyndicated"] = video_syndicated
        if video_type:
            args["videoType"] = video_type
        if page_token:
            args["pageToken"] = page_token

        res_data = self.paged_by_page_token(resource="search", args=args, count=count)

        if return_json:
            return res_data
        else:
            return SearchListResponse.from_dict(res_data)

    def search_by_keywords(
        self,
        *,
        q: Optional[str],
        parts: Optional[Union[str, list, tuple, set]] = None,
        search_type: Optional[Union[str, list, tuple, set]] = None,
        count: Optional[int] = 25,
        limit: Optional[int] = 25,
        page_token: Optional[str] = None,
        return_json: Optional[bool] = False,
        **kwargs: Optional[dict],
    ) -> Union[SearchListResponse, dict]:
        """
        This is simplest usage for search api. You can only passed the keywords to retrieve data from YouTube.
        And the result will include videos,playlists and channels.

        Note: A call to this method has a quota cost of 100 units.

        Args:
            q (str):
                Your keywords can also use the Boolean NOT (-) and OR (|) operators to exclude videos or
                to find videos that are associated with one of several search terms. For example,
                to search for videos matching either "boating" or "sailing",
                set the q parameter value to boating|sailing. Similarly,
                to search for videos matching either "boating" or "sailing" but not "fishing",
                set the q parameter value to boating|sailing -fishing.
                Note that the pipe character must be URL-escaped when it is sent in your API request.
                The URL-escaped value for the pipe character is %7C.
            parts ((str,list,tuple,set) optional):
                The resource parts for you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            search_type ((str,list,tuple,set), optional):
                Parameter restricts a search query to only retrieve a particular type of resource.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
                The default value is video,channel,playlist
                Acceptable values are:
                    - channel
                    - playlist
                    - video
            count (int, optional):
                The count will retrieve videos data.
                Default is 25.
            limit (int, optional):
                The maximum number of items each request retrieve.
                For search, this should not be more than 50.
                Default is 25.
            page_token (str, optional):
                The token of the page of search result to retrieve.
                You can use this retrieve point result page directly.
                And you should know about the the result set for YouTube.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.SearchListResponse instance.
            kwargs:
                If you want use this pass more args. You can use this.

        Returns:
            SearchListResponse or original data
        """
        return self.search(
            parts=parts,
            q=q,
            search_type=search_type,
            count=count,
            limit=limit,
            page_token=page_token,
            return_json=return_json,
            **kwargs,
        )

    def search_by_developer(
        self,
        *,
        parts: Optional[Union[str, list, tuple, set]],
        q: Optional[str] = None,
        count: Optional[int] = 25,
        limit: Optional[int] = 25,
        page_token: Optional[str] = None,
        return_json: Optional[bool] = False,
        **kwargs,
    ) -> Union[SearchListResponse, dict]:
        """
        Parameter restricts the search to only retrieve videos uploaded via the developer's application or website.

        Args:
            parts:
                The resource parts for you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            q:
                Your keywords can also use the Boolean NOT (-) and OR (|) operators to exclude videos or
                to find videos that are associated with one of several search terms. For example,
                to search for videos matching either "boating" or "sailing",
                set the q parameter value to boating|sailing. Similarly,
                to search for videos matching either "boating" or "sailing" but not "fishing",
                set the q parameter value to boating|sailing -fishing.
                Note that the pipe character must be URL-escaped when it is sent in your API request.
                The URL-escaped value for the pipe character is %7C.
            count:
                The count will retrieve videos data.
                Default is 25.
            limit:
                The maximum number of items each request retrieve.
                For search, this should not be more than 50.
                Default is 25.
            page_token:
                The token of the page of search result to retrieve.
                You can use this retrieve point result page directly.
                And you should know about the the result set for YouTube.
            return_json:
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.SearchListResponse instance.
            kwargs:
                If you want use this pass more args. You can use this.

        Returns:
            SearchListResponse or original data
        """
        return self.search(
            for_developer=True,
            search_type="video",
            parts=parts,
            q=q,
            count=count,
            limit=limit,
            page_token=page_token,
            return_json=return_json,
            **kwargs,
        )

    def search_by_mine(
        self,
        *,
        parts: Optional[Union[str, list, tuple, set]],
        q: Optional[str] = None,
        count: Optional[int] = 25,
        limit: Optional[int] = 25,
        page_token: Optional[str] = None,
        return_json: Optional[bool] = False,
        **kwargs,
    ) -> Union[SearchListResponse, dict]:
        """
        Parameter restricts the search to only retrieve videos owned by the authenticated user.

        Note:
            This methods can not use following parameters:
            video_definition, video_dimension, video_duration, video_license,
            video_embeddable, video_syndicated, video_type.
        Args:
            q:
                Your keywords can also use the Boolean NOT (-) and OR (|) operators to exclude videos or
                to find videos that are associated with one of several search terms. For example,
                to search for videos matching either "boating" or "sailing",
                set the q parameter value to boating|sailing. Similarly,
                to search for videos matching either "boating" or "sailing" but not "fishing",
                set the q parameter value to boating|sailing -fishing.
                Note that the pipe character must be URL-escaped when it is sent in your API request.
                The URL-escaped value for the pipe character is %7C.
            parts ((str,list,tuple,set) optional):
                The resource parts for you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            count (int, optional):
                The count will retrieve videos data.
                Default is 25.
            limit (int, optional):
                The maximum number of items each request retrieve.
                For search, this should not be more than 50.
                Default is 25.
            page_token (str, optional):
                The token of the page of search result to retrieve.
                You can use this retrieve point result page directly.
                And you should know about the the result set for YouTube.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.SearchListResponse instance.
            kwargs:
                If you want use this pass more args. You can use this.

        Returns:
            SearchListResponse or original data
        """
        return self.search(
            for_mine=True,
            search_type="video",
            parts=parts,
            q=q,
            count=count,
            limit=limit,
            page_token=page_token,
            return_json=return_json,
            **kwargs,
        )

    def search_by_related_video(
        self,
        *,
        related_to_video_id: str,
        parts: Optional[Union[str, list, tuple, set]] = None,
        region_code: Optional[str] = None,
        relevance_language: Optional[str] = None,
        safe_search: Optional[str] = None,
        count: Optional[int] = 25,
        limit: Optional[int] = 25,
        page_token: Optional[str] = None,
        return_json: Optional[bool] = False,
    ) -> Union[SearchListResponse, dict]:
        """
        Retrieve a list of videos related to that video.

        Args:
            related_to_video_id:
                 A YouTube video ID which result associated with.
            parts:
                The resource parts for you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            region_code:
                Parameter instructs the API to return search results for videos
                that can be viewed in the specified country.
            relevance_language:
                Parameter instructs the API to return search results that are most relevant to the specified language.
            safe_search:
                Parameter indicates whether the search results should include restricted content
                as well as standard content.
                Acceptable values are:
                    - moderate – YouTube will filter some content from search results and, at the least,
                      will filter content that is restricted in your locale. Based on their content,
                      search results could be removed from search results or demoted in search results.
                      This is the default parameter value.
                    - none – YouTube will not filter the search result set.
                    - strict – YouTube will try to exclude all restricted content from the search result set.
                      Based on their content, search results could be removed from search results or
                      demoted in search results.
            count:
                The count will retrieve videos data.
                Default is 25.
            limit:
                The maximum number of items each request retrieve.
                For search, this should not be more than 50.
                Default is 25.
            page_token:
                The token of the page of search result to retrieve.
                You can use this retrieve point result page directly.
                And you should know about the the result set for YouTube.
            return_json:
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.SearchListResponse instance.
        Returns:
            If you want use this pass more args. You can use this.
        """

        return self.search(
            parts=parts,
            related_to_video_id=related_to_video_id,
            search_type="video",
            region_code=region_code,
            relevance_language=relevance_language,
            safe_search=safe_search,
            count=count,
            limit=limit,
            page_token=page_token,
            return_json=return_json,
        )

    def get_subscription_by_id(
        self,
        *,
        subscription_id: Union[str, list, tuple, set],
        parts: Optional[Union[str, list, tuple, set]] = None,
        return_json: Optional[bool] = False,
    ):
        """
        Retrieve subscriptions by given subscription id(s).

        Note:
            This need authorized access token. or you will get no data.

        Args:
            subscription_id ((str,list,tuple,set)):
                The id for subscription that you want to retrieve data.
                You can pass this with single id str, comma-separated id str, or a list,tuple,set of ids.
            parts ((str,list,tuple,set), optional):
                The resource parts for you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.SubscriptionListResponse instance.
        Returns:
            SubscriptionListResponse or original data.
        """

        args = {
            "id": enf_comma_separated(field="subscription_id", value=subscription_id),
            "part": enf_parts(resource="subscriptions", value=parts),
        }

        resp = self._request(resource="subscriptions", method="GET", args=args)
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return SubscriptionListResponse.from_dict(data)

    def get_subscription_by_channel(
        self,
        *,
        channel_id: str,
        parts: Optional[Union[str, list, tuple, set]] = None,
        for_channel_id: Optional[Union[str, list, tuple, set]] = None,
        order: Optional[str] = "relevance",
        count: Optional[int] = 20,
        limit: Optional[int] = 20,
        page_token: Optional[str] = None,
        return_json: Optional[bool] = False,
    ):
        """
        Retrieve the specified channel's subscriptions.

        Note:
             The API returns a 403 (Forbidden) HTTP response code if the specified channel
             does not publicly expose its subscriptions and the request is not authorized
             by the channel's owner.

        Args:
            channel_id (str):
                The id for channel which you want to get subscriptions.
            parts ((str,list,tuple,set) optional):
                The resource parts for subscription you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            for_channel_id ((str,list,tuple,set) optional):
                The parameter specifies a comma-separated list of channel IDs.
                and will then only contain subscriptions matching those channels.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of channel ids.
            order (str, optional):
                The parameter specifies the method that will be used to sort resources in the API response.
                Acceptable values are:
                    alphabetical – Sort alphabetically.
                    relevance – Sort by relevance.
                    unread – Sort by order of activity.
                Default is relevance
            count (int, optional):
                The count will retrieve subscriptions data.
                Default is 20.
                If provide this with None, will retrieve all subscriptions.
            limit (int, optional):
                The maximum number of items each request retrieve.
                For comment threads, this should not be more than 50.
                Default is 20.
            page_token(str, optional):
                The token of the page of subscriptions result to retrieve.
                You can use this retrieve point result page directly.
                And you should know about the the result set for YouTube.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.SubscriptionListResponse instance.
        Returns:
            SubscriptionListResponse or original data.
        """

        if count is None:
            limit = 50  # for subscriptions the max limit for per request is 50
        else:
            limit = min(count, limit)

        args = {
            "channelId": channel_id,
            "part": enf_parts(resource="subscriptions", value=parts),
            "order": order,
            "maxResults": limit,
        }

        if for_channel_id is not None:
            args["forChannelId"] = enf_comma_separated(
                field="for_channel_id", value=for_channel_id
            )

        if page_token is not None:
            args["pageToken"] = page_token

        res_data = self.paged_by_page_token(
            resource="subscriptions", args=args, count=count
        )
        if return_json:
            return res_data
        else:
            return SubscriptionListResponse.from_dict(res_data)

    def get_subscription_by_me(
        self,
        *,
        mine: Optional[bool] = None,
        recent_subscriber: Optional[bool] = None,
        subscriber: Optional[bool] = None,
        parts: Optional[Union[str, list, tuple, set]] = None,
        for_channel_id: Optional[Union[str, list, tuple, set]] = None,
        order: Optional[str] = "relevance",
        count: Optional[int] = 20,
        limit: Optional[int] = 20,
        page_token: Optional[str] = None,
        return_json: Optional[bool] = False,
    ):
        """
        Retrieve your subscriptions.

        Note:
            This can only used in a properly authorized request.
            And for me test the parameter `recent_subscriber` and `subscriber` maybe not working.
            Use the `mine` first.

        Args:
            mine (bool, optional):
                Set this parameter's value to True to retrieve a feed of the authenticated user's subscriptions.
            recent_subscriber (bool, optional):
                Set this parameter's value to true to retrieve a feed of the subscribers of the authenticated user
                in reverse chronological order (newest first).
                And this can only get most recent 1000 subscribers.
            subscriber (bool, optional):
                Set this parameter's value to true to retrieve a feed of the subscribers of
                the authenticated user in no particular order.
            parts ((str,list,tuple,set) optional):
                The resource parts for subscription you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            for_channel_id ((str,list,tuple,set) optional):
                The parameter specifies a comma-separated list of channel IDs.
                and will then only contain subscriptions matching those channels.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of channel ids.
            order (str, optional):
                The parameter specifies the method that will be used to sort resources in the API response.
                Acceptable values are:
                    alphabetical – Sort alphabetically.
                    relevance – Sort by relevance.
                    unread – Sort by order of activity.
                Default is relevance
            count (int, optional):
                The count will retrieve subscriptions data.
                Default is 20.
                If provide this with None, will retrieve all subscriptions.
            limit (int, optional):
                The maximum number of items each request retrieve.
                For subscriptions, this should not be more than 50.
                Default is 20.
            page_token(str, optional):
                The token of the page of subscriptions result to retrieve.
                You can use this retrieve point result page directly.
                And you should know about the the result set for YouTube.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.SubscriptionListResponse instance.

        Returns:
            SubscriptionListResponse or original data.
        """

        if count is None:
            limit = 50  # for subscriptions the max limit for per request is 50
        else:
            limit = min(count, limit)

        args = {
            "part": enf_parts(resource="subscriptions", value=parts),
            "order": order,
            "maxResults": limit,
        }

        if mine is not None:
            args["mine"] = mine
        elif recent_subscriber is not None:
            args["myRecentSubscribers"] = recent_subscriber
        elif subscriber is not None:
            args["mySubscribers"] = subscriber
        else:
            raise PyYouTubeException(
                ErrorMessage(
                    status_code=ErrorCode.MISSING_PARAMS,
                    message=f"Must specify at least one of mine,recent_subscriber,subscriber.",
                )
            )

        if for_channel_id is not None:
            args["forChannelId"] = enf_comma_separated(
                field="for_channel_id", value=for_channel_id
            )

        if page_token is not None:
            args["pageToken"] = page_token

        res_data = self.paged_by_page_token(
            resource="subscriptions", args=args, count=count
        )
        if return_json:
            return res_data
        else:
            return SubscriptionListResponse.from_dict(res_data)

    def get_video_abuse_report_reason(
        self,
        *,
        parts: Optional[Union[str, list, tuple, set]] = None,
        hl: Optional[str] = "en_US",
        return_json: Optional[bool] = False,
    ) -> Union[VideoAbuseReportReasonListResponse, dict]:
        """
        Retrieve a list of reasons that can be used to report abusive videos.

        Notes:
            This requires your authorization.

        Args:
            parts:
                The resource parts for abuse reason you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            hl:
                If provide this. Will return report reason's language localized info.
                This value need https://developers.google.com/youtube/v3/docs/i18nLanguages.
            return_json:
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.VideoAbuseReportReasonListResponse instance.
        Returns:
            VideoAbuseReportReasonListResponse or original data.
        """

        args = {
            "part": enf_parts(resource="videoAbuseReportReasons", value=parts),
            "hl": hl,
        }

        resp = self._request(resource="videoAbuseReportReasons", args=args)
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return VideoAbuseReportReasonListResponse.from_dict(data)

    def get_video_categories(
        self,
        *,
        category_id: Optional[Union[str, list, tuple, set]] = None,
        region_code: Optional[str] = None,
        parts: Optional[Union[str, list, tuple, set]] = None,
        hl: Optional[str] = "en_US",
        return_json: Optional[bool] = False,
    ):
        """
        Retrieve video categories by category id or region code.

        Args:
            category_id ((str,list,tuple,set), optional):
                The id for video category thread that you want to retrieve data.
                You can pass this with single id str, comma-separated id str, or a list,tuple,set of ids.
            region_code (str, optional):
                The region code that you want to retrieve guide categories.
                The parameter value is an ISO 3166-1 alpha-2 country code.
                Refer: https://www.iso.org/iso-3166-country-codes.html
            parts ((str,list,tuple,set) optional):
                The resource parts for you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            hl (str, optional):
                If provide this. Will return video category's language localized info.
                This value need https://developers.google.com/youtube/v3/docs/i18nLanguages.
                Default is en_US.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.VideoCategoryListResponse instance.
        Returns:
            VideoCategoryListResponse or original data
        """
        args = {
            "part": enf_parts(resource="videoCategories", value=parts),
            "hl": hl,
        }

        if category_id is not None:
            args["id"] = enf_comma_separated(field="category_id", value=category_id)
        elif region_code is not None:
            args["regionCode"] = region_code
        else:
            raise PyYouTubeException(
                ErrorMessage(
                    status_code=ErrorCode.MISSING_PARAMS,
                    message="Specify at least one of category_id or region_code",
                )
            )

        resp = self._request(resource="videoCategories", method="GET", args=args)
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return VideoCategoryListResponse.from_dict(data)

    def get_video_by_id(
        self,
        *,
        video_id: Union[str, list, tuple, set],
        parts: Optional[Union[str, list, tuple, set]] = None,
        hl: Optional[str] = "en_US",
        max_height: Optional[int] = None,
        max_width: Optional[int] = None,
        return_json: Optional[bool] = False,
        proxy:Optional[str]=None
    ):
        """
        Retrieve video data by given video id.

        Args:
            video_id ((str,list,tuple,set)):
                The id for video that you want to retrieve data.
                You can pass this with single id str, comma-separated id str, or a list,tuple,set of ids.
            parts ((str,list,tuple,set), optional):
                The resource parts for you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            hl (str, optional):
                If provide this. Will return video's language localized info.
                This value need https://developers.google.com/youtube/v3/docs/i18nLanguages.
            max_height (int, optional):
                Specifies the maximum height of the embedded player returned in the player.embedHtml property.
                Acceptable values are 72 to 8192, inclusive.
            max_width (int, optional):
                Specifies the maximum width of the embedded player returned in the player.embedHtml property.
                Acceptable values are 72 to 8192, inclusive.
                If provide max_height at the same time. This will may be shorter than max_height.
                For more https://developers.google.com/youtube/v3/docs/videos/list#parameters.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.VideoListResponse instance.

        Returns:
            VideoListResponse or original data
        """

        args = {
            "id": enf_comma_separated(field="video_id", value=video_id),
            "part": enf_parts(resource="videos", value=parts),
            "hl": hl,
        }
        if max_height is not None:
            args["maxHeight"] = max_height
        if max_width is not None:
            args["maxWidth"] = max_width

        resp = self._request(resource="videos", method="GET", args=args)
        data = self._parse_response(resp)

        if return_json:
            return data
        else:
            return VideoListResponse.from_dict(data)

    def get_videos_by_chart(
        self,
        *,
        chart: str,
        parts: Optional[Union[str, list, tuple, set]] = None,
        hl: Optional[str] = "en_US",
        max_height: Optional[int] = None,
        max_width: Optional[int] = None,
        region_code: Optional[str] = None,
        category_id: Optional[str] = "0",
        count: Optional[int] = 5,
        limit: Optional[int] = 5,
        page_token: Optional[str] = None,
        return_json: Optional[bool] = False,
    ):
        """
        Retrieve a list of YouTube's most popular videos.

        Args:
            chart (str):
                The chart string for you want to retrieve data.
                Acceptable values are: mostPopular
            parts ((str,list,tuple,set), optional):
                The resource parts for you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            hl (str, optional):
                If provide this. Will return playlist's language localized info.
                This value need https://developers.google.com/youtube/v3/docs/i18nLanguages.
            max_height (int, optional):
                Specifies the maximum height of the embedded player returned in the player.embedHtml property.
                Acceptable values are 72 to 8192, inclusive.
            max_width (int, optional):
                Specifies the maximum width of the embedded player returned in the player.embedHtml property.
                Acceptable values are 72 to 8192, inclusive.
                If provide max_height at the same time. This will may be shorter than max_height.
                For more https://developers.google.com/youtube/v3/docs/videos/list#parameters.
            region_code (str, optional):
                This parameter instructs the API to select a video chart available in the specified region.
                Value is an ISO 3166-1 alpha-2 country code.
            category_id (str, optional):
                The id for video category that you want to filter.
                Default is 0.
            count (int, optional):
                The count will retrieve videos data.
                Default is 5.
                If provide this with None, will retrieve all videos.
            limit (int, optional):
                The maximum number of items each request retrieve.
                For videos, this should not be more than 50.
                Default is 5.
            page_token(str, optional):
                The token of the page of videos result to retrieve.
                You can use this retrieve point result page directly.
                And you should know about the the result set for YouTube.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.PlaylistListResponse instance.

        Returns:
            VideoListResponse or original data
        """

        if count is None:
            limit = 50  # for videos the max limit for per request is 50
        else:
            limit = min(count, limit)

        args = {
            "chart": chart,
            "part": enf_parts(resource="videos", value=parts),
            "hl": hl,
            "maxResults": limit,
            "videoCategoryId": category_id,
        }
        if max_height is not None:
            args["maxHeight"] = max_height
        if max_width is not None:
            args["maxWidth"] = max_width
        if region_code:
            args["regionCode"] = region_code

        if page_token is not None:
            args["pageToken"] = page_token

        res_data = self.paged_by_page_token(resource="videos", args=args, count=count)
        if return_json:
            return res_data
        else:
            return VideoListResponse.from_dict(res_data)

    def get_videos_by_myrating(
        self,
        *,
        rating: str,
        parts: Optional[Union[str, list, tuple, set]] = None,
        hl: Optional[str] = "en_US",
        max_height: Optional[int] = None,
        max_width: Optional[int] = None,
        count: Optional[int] = 5,
        limit: Optional[int] = 5,
        page_token: Optional[str] = None,
        return_json: Optional[bool] = False,
    ):
        """
        Retrieve video data by my ration.

        Args:
            rating (str):
                The rating string for you to retrieve data.
                Acceptable values are: dislike, like
            parts ((str,list,tuple,set), optional):
                The resource parts for you want to retrieve.
                If not provide, use default public parts.
                You can pass this with single part str, comma-separated parts str or a list,tuple,set of parts.
            hl (str, optional):
                If provide this. Will return video's language localized info.
                This value need https://developers.google.com/youtube/v3/docs/i18nLanguages.
            max_height (int, optional):
                Specifies the maximum height of the embedded player returned in the player.embedHtml property.
                Acceptable values are 72 to 8192, inclusive.
            max_width (int, optional):
                Specifies the maximum width of the embedded player returned in the player.embedHtml property.
                Acceptable values are 72 to 8192, inclusive.
                If provide max_height at the same time. This will may be shorter than max_height.
                For more https://developers.google.com/youtube/v3/docs/videos/list#parameters.
            count (int, optional):
                The count will retrieve videos data.
                Default is 5.
                If provide this with None, will retrieve all videos.
            limit (int, optional):
                The maximum number of items each request retrieve.
                For videos, this should not be more than 50.
                Default is 5.
            page_token(str, optional):
                The token of the page of videos result to retrieve.
                You can use this retrieve point result page directly.
                And you should know about the the result set for YouTube.
            return_json(bool, optional):
                The return data type. If you set True JSON data will be returned.
                False will return a pyyoutube.VideoListResponse instance.
        Returns:
            VideoListResponse or original data
        """

        if self._access_token is None:
            raise PyYouTubeException(
                ErrorMessage(
                    status_code=ErrorCode.NEED_AUTHORIZATION,
                    message="This method can only used with authorization",
                )
            )

        if count is None:
            limit = 50  # for videos the max limit for per request is 50
        else:
            limit = min(count, limit)

        args = {
            "myRating": rating,
            "part": enf_parts(resource="videos", value=parts),
            "hl": hl,
            "maxResults": limit,
        }

        if max_height is not None:
            args["maxHeight"] = max_height
        if max_width is not None:
            args["maxWidth"] = max_width

        if page_token is not None:
            args["pageToken"] = page_token

        res_data = self.paged_by_page_token(resource="videos", args=args, count=count)
        if return_json:
            return res_data
        else:
            return VideoListResponse.from_dict(res_data)



"""
    These are some mixin for models
"""
import datetime
from typing import Optional

import isodate
from isodate.isoerror import ISO8601Error

from pyyoutube.error import ErrorCode, ErrorMessage, PyYouTubeException


class DatetimeTimeMixin:
    @staticmethod
    def string_to_datetime(dt_str: Optional[str]) -> Optional[datetime.datetime]:
        """
        Convert datetime string to datetime instance.
        original string format is YYYY-MM-DDThh:mm:ss.sZ.
        :return:
        """
        if not dt_str:
            return None
        try:
            r = isodate.parse_datetime(dt_str)
        except ISO8601Error as e:
            raise PyYouTubeException(
                ErrorMessage(status_code=ErrorCode.INVALID_PARAMS, message=e.args[0])
            )
        else:
            return r



import base64
from selenium.webdriver.support.wait import WebDriverWait
from seleniumwire import webdriver as wired_webdriver
from selenium import webdriver
from bs4 import BeautifulSoup
from selenium.webdriver.common.by import By
from selenium.webdriver.chrome.service import Service
from webdriver_manager.chrome import ChromeDriverManager
from selenium.webdriver.support import expected_conditions as EC
from seleniumwire import webdriver
from ytvs.embedding_iframe_api_creator import EmbeddingIFrameAPICreator


class EmbeddingPermissionChecker:

    def __init__(self):
        self._embedding_url_creator = EmbeddingIFrameAPICreator()

    def is_embeddable(self, video_id: str):

        iframe_api = self._embedding_url_creator.create(video_id)
        encoded_html = base64.b64encode(iframe_api.encode('utf-8'))
        data_url = f"data:text/html;base64,{encoded_html.decode('utf-8')}"

        options = webdriver.ChromeOptions()
        options.add_argument('--disable-dev-shm-usage')
        is_headless = True
        if is_headless is True:
            options.add_argument('headless')
            options.add_argument('--no-sandbox')

        driver = wired_webdriver.Chrome(service=Service(ChromeDriverManager().install()), options=options)
        driver.get(data_url)
        wait = WebDriverWait(driver, 100)
        iframe = wait.until(EC.visibility_of_element_located((By.TAG_NAME, "iframe")))

        # iframe 요소 찾기 및 전환
        driver.switch_to.frame(iframe)
        iframe_html = driver.page_source
        soup = BeautifulSoup(iframe_html, "html.parser", from_encoding="utf-8")
        error_container_elem = soup.find("div", {'class': 'ytp-error'})

        if error_container_elem is not None:
            return False
        return True





"""
    Subscription resource implementation.
"""

from typing import Optional, Union

from pyyoutube.error import PyYouTubeException, ErrorCode, ErrorMessage
from pyyoutube.resources.base_resource import Resource
from pyyoutube.models import Subscription, SubscriptionListResponse
from pyyoutube.utils.params_checker import enf_parts, enf_comma_separated


class SubscriptionsResource(Resource):
    """A subscription resource contains information about a YouTube user subscription.

    References: https://developers.google.com/youtube/v3/docs/subscriptions
    """

    def list(
        self,
        parts: Optional[Union[str, list, tuple, set]] = None,
        channel_id: Optional[str] = None,
        subscription_id: Optional[Union[str, list, tuple, set]] = None,
        mine: Optional[bool] = None,
        my_recent_subscribers: Optional[bool] = None,
        my_subscribers: Optional[bool] = None,
        for_channel_id: Optional[Union[str, list, tuple, set]] = None,
        max_results: Optional[int] = None,
        on_behalf_of_content_owner: Optional[str] = None,
        on_behalf_of_content_owner_channel: Optional[str] = None,
        order: Optional[str] = None,
        page_token: Optional[str] = None,
        return_json: bool = False,
        **kwargs: Optional[dict],
    ) -> Union[dict, SubscriptionListResponse]:
        """Returns subscription resources that match the API request criteria.

        Args:
            parts:
                Comma-separated list of one or more channel resource properties.
                Accepted values: id,contentDetails,snippet,subscriberSnippet
            channel_id:
                Specifies a YouTube channel ID. The API will only return that channel's subscriptions.
            subscription_id:
                Specifies a comma-separated list of the YouTube subscription ID(s) for the resource(s)
                that are being retrieved.
            mine:
                Set this parameter's value to true to retrieve a feed of the authenticated user's subscriptions.
            my_recent_subscribers:
                Set this parameter's value to true to retrieve a feed of the subscribers of the authenticated user
                in reverse chronological order (the newest first).
            my_subscribers:
                Set this parameter's value to true to retrieve a feed of the subscribers of the authenticated user
                in no particular order.
            for_channel_id:
                Specifies a comma-separated list of channel IDs.
                The API response will then only contain subscriptions matching those channels.
            max_results:
                The parameter specifies the maximum number of items that should be returned
                the result set.
                Acceptable values are 0 to 50, inclusive. The default value is 5.
            on_behalf_of_content_owner:
                The onBehalfOfContentOwner parameter indicates that the request's authorization
                credentials identify a YouTube CMS user who is acting on behalf of the content
                owner specified in the parameter value. This parameter is intended for YouTube
                content partners that own and manage many difference YouTube channels. It allows
                content owners to authenticate once and get access to all their video and channel
                data, without having to provide authentication credentials for each individual channel.
                The CMS account that the user authenticates with must be linked to the specified YouTube content owner.
            on_behalf_of_content_owner_channel:
                The onBehalfOfContentOwnerChannel parameter specifies the YouTube channel ID of
                the channel to which a video is being added. This parameter is required when a request
                specifies a value for the onBehalfOfContentOwner parameter, and it can only be used in
                conjunction with that parameter. In addition, the request must be authorized using a
                CMS account that is linked to the content owner that the onBehalfOfContentOwner parameter
                specifies. Finally, the channel that the onBehalfOfContentOwnerChannel parameter value
                specifies must be linked to the content owner that the onBehalfOfContentOwner parameter specifies.
            order:
                Specifies the method that will be used to sort resources in the API response.
                Acceptable values are:
                    - alphabetical: Sort alphabetically.
                    - relevance: Sort by relevance. Default.
                    - unread: Sort by order of activity.
            page_token:
                The parameter identifies a specific page in the result set that should be returned.
            return_json:
                Type for returned data. If you set True JSON data will be returned.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Subscriptions data.
        Raises:
            PyYouTubeException: Missing filter parameter.
        """

        params = {
            "part": enf_parts(resource="subscriptions", value=parts),
            "forChannelId": enf_comma_separated(
                field="for_channel_id", value=for_channel_id
            ),
            "maxResults": max_results,
            "onBehalfOfContentOwner": on_behalf_of_content_owner,
            "onBehalfOfContentOwnerChannel": on_behalf_of_content_owner_channel,
            "order": order,
            "pageToken": page_token,
            **kwargs,
        }

        if channel_id is not None:
            params["channelId"] = channel_id
        elif subscription_id is not None:
            params["id"] = enf_comma_separated(
                field="subscription_id", value=subscription_id
            )
        elif mine is not None:
            params["mine"] = mine
        elif my_recent_subscribers is not None:
            params["myRecentSubscribers"] = my_recent_subscribers
        elif my_subscribers is not None:
            params["mySubscribers"] = my_subscribers
        else:
            raise PyYouTubeException(
                ErrorMessage(
                    status_code=ErrorCode.MISSING_PARAMS,
                    message=f"Specify at least one of channel_id,subscription_id,mine,my_recent_subscribers or mySubscribers",
                )
            )

        response = self._client.request(path="subscriptions", params=params)
        data = self._client.parse_response(response=response)
        return data if return_json else SubscriptionListResponse.from_dict(data)

    def insert(
        self,
        body: Union[dict, Subscription],
        parts: Optional[Union[str, list, tuple, set]] = None,
        return_json: bool = False,
        **kwargs: Optional[dict],
    ) -> Union[dict, Subscription]:
        """Adds a subscription for the authenticated user's channel.

        Args:
            body:
                Provide subscription data in the request body. You can give dataclass or just a dict with data.
            parts:
                The part parameter serves two purposes in this operation. It identifies the properties
                that the write operation will set as well as the properties that the API response will include.
                Accepted values: id,contentDetails,snippet,subscriberSnippet
            return_json:
                Type for returned data. If you set True JSON data will be returned.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Subscription data

        """
        params = {
            "part": enf_parts(resource="subscriptions", value=parts),
            **kwargs,
        }
        response = self._client.request(
            method="POST",
            path="subscriptions",
            params=params,
            json=body,
        )
        data = self._client.parse_response(response=response)
        return data if return_json else Subscription.from_dict(data)

    def delete(
        self,
        subscription_id: str,
        **kwargs: Optional[dict],
    ) -> bool:
        """Deletes a subscription.

        Args:
            subscription_id:
                Specifies the YouTube subscription ID for the resource that is being deleted.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Subscription delete status.
        """
        params = {
            "id": subscription_id,
            **kwargs,
        }
        response = self._client.request(
            method="DELETE", path="subscriptions", params=params
        )
        if response.ok:
            return True
        self._client.parse_response(response=response)



from math import ceil
import base64
import io
import traceback


def byte(n):
    return bytes((n,))


def varint_encode(offset):
    '''In this encoding system, for each 8-bit byte, the first bit is 1 if there are more bytes, and 0 is this is the last one.
    The next 7 bits are data. These 7-bit sections represent the data in Little endian order. For example, suppose the data is
    aaaaaaabbbbbbbccccccc (each of these sections is 7 bits). It will be encoded as:
    1ccccccc 1bbbbbbb 0aaaaaaa

    This encoding is used in youtube parameters to encode offsets and to encode the length for length-prefixed data.
    See https://developers.google.com/protocol-buffers/docs/encoding#varints for more info.'''
    needed_bytes = ceil(offset.bit_length() / 7) or 1  # (0).bit_length() returns 0, but we need 1 in that case.
    encoded_bytes = bytearray(needed_bytes)
    for i in range(0, needed_bytes - 1):
        encoded_bytes[i] = (offset & 127) | 128  # 7 least significant bits
        offset = offset >> 7
    encoded_bytes[-1] = offset & 127  # leave first bit as zero for last byte

    return bytes(encoded_bytes)


def varint_decode(encoded):
    decoded = 0
    for i, byte in enumerate(encoded):
        decoded |= (byte & 127) << 7 * i

        if not (byte & 128):
            break
    return decoded


def string(field_number, data):
    data = as_bytes(data)
    return _proto_field(2, field_number, varint_encode(len(data)) + data)


nested = string


def uint(field_number, value):
    return _proto_field(0, field_number, varint_encode(value))


def _proto_field(wire_type, field_number, data):
    ''' See https://developers.google.com/protocol-buffers/docs/encoding#structure '''
    return varint_encode((field_number << 3) | wire_type) + data


def percent_b64encode(data):
    return base64.urlsafe_b64encode(data).replace(b'=', b'%3D')


def unpadded_b64encode(data):
    return base64.urlsafe_b64encode(data).replace(b'=', b'')


def as_bytes(value):
    if isinstance(value, str):
        return value.encode('utf-8')
    return value


def read_varint(data):
    result = 0
    i = 0
    while True:
        try:
            byte = data.read(1)[0]
        except IndexError:
            if i == 0:
                raise EOFError()
            raise Exception('Unterminated varint starting at ' + str(data.tell() - i))
        result |= (byte & 127) << 7 * i
        if not byte & 128:
            break

        i += 1
    return result


def read_group(data, end_sequence):
    start = data.tell()
    index = data.original.find(end_sequence, start)
    if index == -1:
        raise Exception('Unterminated group')
    data.seek(index + len(end_sequence))
    return data.original[start:index]


def read_protobuf(data):
    data_original = data
    data = io.BytesIO(data)
    data.original = data_original
    while True:
        try:
            tag = read_varint(data)
        except EOFError:
            break
        wire_type = tag & 7
        field_number = tag >> 3

        if wire_type == 0:
            value = read_varint(data)
        elif wire_type == 1:
            value = data.read(8)
        elif wire_type == 2:
            length = read_varint(data)
            value = data.read(length)
        elif wire_type == 3:
            end_bytes = varint_encode((field_number << 3) | 4)
            value = read_group(data, end_bytes)
        elif wire_type == 5:
            value = data.read(4)
        else:
            raise Exception("Unknown wire type: " + str(wire_type) + ", at position " + str(data.tell()))
        yield (wire_type, field_number, value)


def parse(data, include_wire_type=False):
    '''Returns a dict mapping field numbers to values

    data is the protobuf structure, which must not be b64-encoded'''
    if include_wire_type:
        return {field_number: [wire_type, value]
                for wire_type, field_number, value in read_protobuf(data)}
    return {field_number: value
            for _, field_number, value in read_protobuf(data)}


base64_enc_funcs = {
    'base64': base64.urlsafe_b64encode,
    'base64s': unpadded_b64encode,
    'base64p': percent_b64encode,
}


def _make_protobuf(data):
    # must be dict mapping field_number to [wire_type, value]
    if isinstance(data, dict):
        new_data = []
        for field_num, (wire_type, value) in sorted(data.items()):
            new_data.append((wire_type, field_num, value))
        data = new_data
    if isinstance(data, str):
        return data.encode('utf-8')
    elif len(data) == 2 and data[0] in list(base64_enc_funcs.keys()):
        return base64_enc_funcs[data[0]](_make_protobuf(data[1]))
    elif isinstance(data, list):
        result = b''
        for field in data:
            if field[0] == 0:
                result += uint(field[1], field[2])
            elif field[0] == 2:
                result += string(field[1], _make_protobuf(field[2]))
            else:
                raise NotImplementedError('Wire type ' + str(field[0])
                                          + ' not implemented')
        return result
    return data


def make_protobuf(data):
    return _make_protobuf(data).decode('ascii')


def _set_protobuf_value(data, *path, value):
    if not path:
        return value
    op = path[0]
    if op in base64_enc_funcs:
        inner_data = b64_to_bytes(data)
        return base64_enc_funcs[op](
            _set_protobuf_value(inner_data, *path[1:], value=value)
        )
    pb_dict = parse(data, include_wire_type=True)
    pb_dict[op][1] = _set_protobuf_value(
        pb_dict[op][1], *path[1:], value=value
    )
    return _make_protobuf(pb_dict)


def set_protobuf_value(data, *path, value):
    '''Set a field's value in a raw protobuf structure

    path is a list of field numbers and/or base64 encoding directives

    The directives are
        base64: normal base64 encoding with equal signs padding
        base64s ("stripped"): no padding
        base64p: %3D instead of = for padding

    return new_protobuf, err'''
    try:
        new_protobuf = _set_protobuf_value(data, *path, value=value)
        return new_protobuf.decode('ascii'), None
    except Exception:
        return None, traceback.format_exc()


def b64_to_bytes(data):
    if isinstance(data, bytes):
        data = data.decode('ascii')
    data = data.replace("%3D", "=")
    return base64.urlsafe_b64decode(data + "=" * ((4 - len(data) % 4) % 4))



import re
from difflib import SequenceMatcher

import requests
from bs4 import BeautifulSoup
from selenium.webdriver.support.wait import WebDriverWait
from seleniumwire import webdriver
from webdriver_manager.chrome import ChromeDriverManager
from selenium.common import NoSuchElementException
from seleniumwire import webdriver as wired_webdriver
from selenium import webdriver
from seleniumwire import webdriver, utils
import requests
from bs4 import BeautifulSoup
from selenium.webdriver.common.by import By
from selenium.webdriver.chrome.service import Service
from webdriver_manager.chrome import ChromeDriverManager
from selenium.webdriver.support import expected_conditions as EC

from ytvs.client.unofficial.resources import SearchResource
from ytvs.client.unofficial.resources.base_resource import Resource
from ytvs.embedding_permission_checker import EmbeddingPermissionChecker
from ytvs.extractor import YoutubeBasicInfoExtractor
from ytvs.filter import SearchType, Order, VideoDuration
from ytvs.utils import try_get, str_to_int, _search_attribute, _search_attributes


def similar(a, b):
    return SequenceMatcher(None, a, b).ratio()


class VideoResource(Resource):
    KIND = 'youtube#video'

    def get_video_by_id(self, video_id: str):
        """
        주어진 비디오 ID에 대한 YouTube 비디오 정보를 조회.
        :param video_id: 정보를 가져올 비디오의 ID.
        :return: 파싱된 비디오 정보.
        """
        # 1. 영상 페이지 요청
        data = self._request_data(video_id)

        # 2. 영상 페이지에서 데이터 추출
        yt_initial_data = YoutubeBasicInfoExtractor.extract_yt_initial_data(data)

        # 3. 데이터 분석 및 파싱
        parsed_data = self._parse_video_data_from_initial_data(yt_initial_data)
        parsed_data_from_html = self._parsed_video_data_from_html_txt(data)

        embeddable_checker = EmbeddingPermissionChecker()
        # is_embeddable = embeddable_checker.is_embeddable(video_id)

        # 4. 추가 데이터 병합
        parsed_data['items'][0]['id'] = video_id
        parsed_data['items'][0]['snippet']['publishedAt'] = parsed_data_from_html['published_date']
        parsed_data['items'][0]['status'] = {}
        # parsed_data['items'][0]['status']['embeddable'] = is_embeddable
        #
        # title = parsed_data['items'][0]['snippet']['title']
        # thumbnails_and_duration = self._search_video_thumbnails_and_duration(title)
        #
        # parsed_data['items'][0]['snippet']['thumbnails'] = thumbnails_and_duration['thumbnails']
        # parsed_data['items'][0]['contentDetails'] = {}
        # parsed_data['items'][0]['contentDetails']['duration'] = thumbnails_and_duration['duration']

        # 5. 반환
        return parsed_data

    def _search_video_thumbnails_and_duration(self, title):
        # @NOTE: Youtube 검색 API를 이용시, '#' 문자가 있으면 올바르게 검색되지 않는 이슈로 제거함
        # 하지만, # 문자가 포함되어있는경우, 검색이 안되는 정확한 원인파악이 필요
        # -> Youtube 페이지에서는 #이 있어도 검색이 정확하게됨.(API 쿼리 구문에 # 이포함됨. 차이가 뭐지)

        refined_title = title.replace("#", "")
        search_resource = SearchResource(self._client)
        searched = search_resource.search(q=refined_title, search_type=SearchType.VIDEO.value,
                                          order=Order.RELEVANCE.value,
                                          video_duration=VideoDuration.ANY.value)
        searched_item = self._find_by_title_in_search_list(title, searched['items'])
        if searched_item is not None:
            duration = searched['items'][0]['snippet']['duration']
            thumbnails = searched['items'][0]['snippet']['thumbnails']
        else:
            duration = None
            thumbnails = None
        return {
            'duration': duration,
            'thumbnails': thumbnails
        }

    def _find_by_title_in_search_list(self, title: str, searches):
        # 제목이 가장 유사한 녀석을 찾음
        most_similarity = 0
        most_similarity_idx = 0
        for i, searched_item in enumerate(searches[:5]):
            searched_item_title = searched_item['snippet']['title']
            similarity = similar(searched_item_title, title)

            if similarity > most_similarity:
                most_similarity = similarity
                most_similarity_idx = i
        if most_similarity > 0.7:
            return searches[most_similarity_idx]
        else:
            return None

    def _request_data(self, video_id: str):
        """
        YouTube 비디오 페이지의 HTML 데이터를 요청.
        :param video_id: 요청할 비디오의 ID.
        :return: 비디오 페이지의 HTML 텍스트를 반환.
        """
        # 1. Video resource 요청 URL 생성
        url = self.build_video_resource_url(video_id)

        # # 2. Video 페이지 요청 및 데이터 추출
        return self._client.fetch_url(url).decode("utf-8")

    def _parsed_video_data_from_html_txt(self, html_text):
        # *******
        soup = BeautifulSoup(html_text, "html.parser", from_encoding="utf-8")
        # date published
        published_date_text = soup.find("meta", itemprop="datePublished")['content']
        # get the duration of the video
        # 광고 영상때문에, HTML 페이지에서는 확실한 영상의 재생시간 획득을 보장할수가 없음...
        # video_duration_text = soup.find("span", {"class": "ytp-time-duration"}).text
        return {'published_date': published_date_text}

    def _parse_video_data_from_initial_data(self, initial_data):
        """
        초기 데이터에서 YouTube 비디오 데이터를 파싱.
        :param initial_data: YouTube 비디오 페이지에서 추출된 초기 데이터.
        :return: 파싱된 비디오 데이터를 반환.
        """
        # 1. 데이터 분석 및 파싱
        contents = try_get(initial_data,
                           lambda x: x['init_data']['contents']['twoColumnWatchNextResults']['results']['results'][
                               'contents'], list) or []

        # 2. Content 획득
        video_primary_video_renderer = _search_attribute(contents, "videoPrimaryInfoRenderer")
        if video_primary_video_renderer:
            video = YoutubeBasicInfoExtractor.extract_video(video_primary_video_renderer)
        else:
            video = {}
        video_secondary_info_renderer = _search_attribute(contents, "videoSecondaryInfoRenderer")
        if video_secondary_info_renderer:
            description = try_get(video_secondary_info_renderer,
                                  lambda x: x['attributedDescription']['content'], str)
            owner = try_get(video_secondary_info_renderer,
                            lambda x: x['owner']['videoOwnerRenderer'], dict)
            owner = YoutubeBasicInfoExtractor.extract_video_owner_renderer(owner)
        else:
            description = None
            owner = {}

        item_section_renderers = _search_attributes(contents, 'itemSectionRenderer')
        comments_header_renderer = _search_attribute(item_section_renderers, "commentsEntryPointHeaderRenderer",
                                                     max_depth=6)
        comments_header = YoutubeBasicInfoExtractor.extract_comments_header_renderer(comments_header_renderer)

        return {
            'etag': None,
            'kind': 'youtube#videoListResponse',
            'items': [{
                'kind': self.KIND,
                'localizations': None,
                'etag': None,
                'snippet': {
                    'title': try_get(video, lambda x: x['title'], str),
                    'categoryId': None,
                    'channelId': try_get(owner, lambda x: x['channel_id'], str),
                    'channelTitle': try_get(owner, lambda x: x['title'], str),
                    'description': description,
                    'liveBroadcastContent': None,
                    'localized': None,
                    'publishedAt': try_get(video, lambda x: x['published_at'], str),
                    'tags': [],
                    'thumbnails': None,
                    'hashtags': try_get(video, lambda x: x['super_title_links'], list)
                },
                'statistics': {
                    'dislikeCount': None,
                    'commentCount': try_get(comments_header, lambda x: x['comment_count']),
                    'likeCount': try_get(video, lambda x: x['like_count']),
                    'viewCount': try_get(video, lambda x: x['view_count']),
                }
            }],
            'nextPageToken': None,
            'prevPageToken': None,
            'pageInfo': {
                'totalResults': 1,
                'resultsPerPage': 1
            }
        }



from datetime import datetime
# import settings
import socks
import sockshandler
import gzip
try:
    import brotli
    have_brotli = True
except ImportError:
    have_brotli = False
import urllib.parse
import re
import time
import os
import json
import gevent
import gevent.queue
import gevent.lock
import collections
import stem
import stem.control
import traceback

# The trouble with the requests library: It ships its own certificate bundle via certifi
#  instead of using the system certificate store, meaning self-signed certificates
#  configured by the user will not work. Some draconian networks block TLS unless a corporate
#  certificate is installed on the system. Additionally, some users install a self signed cert
#  in order to use programs to modify or monitor requests made by programs on the system.

# Finally, certificates expire and need to be updated, or are sometimes revoked. Sometimes
#  certificate authorites go rogue and need to be untrusted. Since we are going through Tor exit nodes,
#  this becomes all the more important. A rogue CA could issue a fake certificate for accounts.google.com, and a
#  malicious exit node could use this to decrypt traffic when logging in and retrieve passwords. Examples:
#   https://www.engadget.com/2015/10/29/google-warns-symantec-over-certificates/
#   https://nakedsecurity.sophos.com/2013/12/09/serious-security-google-finds-fake-but-trusted-ssl-certificates-for-its-domains-made-in-france/

# In the requests documentation it says:
#    "Before version 2.16, Requests bundled a set of root CAs that it trusted, sourced from the Mozilla trust store.
#     The certificates were only updated once for each Requests version. When certifi was not installed,
#     this led to extremely out-of-date certificate bundles when using significantly older versions of Requests.
#     For the sake of security we recommend upgrading certifi frequently!"
#   (http://docs.python-requests.org/en/master/user/advanced/#ca-certificates)

# Expecting users to remember to manually update certifi on Linux isn't reasonable in my view.
#  On windows, this is even worse since I am distributing all dependencies. This program is not
#  updated frequently, and using requests would lead to outdated certificates. Certificates
#  should be updated with OS updates, instead of thousands of developers of different programs
#  being expected to do this correctly 100% of the time.

# There is hope that this might be fixed eventually:
#   https://github.com/kennethreitz/requests/issues/2966

# Until then, I will use a mix of urllib3 and urllib.
import urllib3
import urllib3.contrib.socks

URL_ORIGIN = "/https://www.youtube.com"

connection_pool = urllib3.PoolManager(cert_reqs='CERT_REQUIRED')


class TorManager:
    MAX_TRIES = 3
    # Remember the 7-sec wait times, so make cooldown be two of those
    # (otherwise it will retry forever if 429s never end)
    COOLDOWN_TIME = 14

    def __init__(self, tor_port:int=9050):
        self.tor_port = tor_port
        self.old_tor_connection_pool = None
        self.tor_connection_pool = urllib3.contrib.socks.SOCKSProxyManager(
            #'socks5h://127.0.0.1:' + str(settings.tor_port) + '/',
            'socks5h://127.0.0.1:' + str(tor_port) + '/',
            cert_reqs='CERT_REQUIRED')
        self.tor_pool_refresh_time = time.monotonic()
        # settings.add_setting_changed_hook(
        #     'tor_port',
        #     lambda old_val, new_val: self.refresh_tor_connection_pool(),
        # )

        self.new_identity_lock = gevent.lock.BoundedSemaphore(1)
        self.last_new_identity_time = time.monotonic() - 20
        self.try_num = 1

    def refresh_tor_connection_pool(self):
        self.tor_connection_pool.clear()

        # Keep a reference for 5 min to avoid it getting garbage collected
        # while sockets still in use
        self.old_tor_connection_pool = self.tor_connection_pool

        self.tor_connection_pool = urllib3.contrib.socks.SOCKSProxyManager(
            #'socks5h://127.0.0.1:' + str(settings.tor_port) + '/',
            'socks5h://127.0.0.1:' + str(self.tor_port) + '/',
            cert_reqs='CERT_REQUIRED')
        self.tor_pool_refresh_time = time.monotonic()

    def get_tor_connection_pool(self):
        # Tor changes circuits after 10 minutes:
        # https://tor.stackexchange.com/questions/262/for-how-long-does-a-circuit-stay-alive
        current_time = time.monotonic()

        # close pool after 5 minutes
        if current_time - self.tor_pool_refresh_time > 300:
            self.refresh_tor_connection_pool()

        return self.tor_connection_pool

    def new_identity(self, time_failed_request_started):
        '''return error, or None if no error and the identity is fresh'''

        # The overall pattern at maximum (always returning 429) will be
        # R N (0) R N (6) R N (6) R | (12) R N (0) R N (6) ...
        # where R is a request, N is a new identity, (x) is a wait time of
        # x sec, and | is where we give up and display an error to the user.

        print('new_identity: new_identity called')
        # blocks if another greenlet currently has the lock
        self.new_identity_lock.acquire()
        print('new_identity: New identity lock acquired')

        try:
            # This was caused by a request that failed within a previous,
            # stale identity
            if time_failed_request_started <= self.last_new_identity_time:
                print('new_identity: Cancelling; request was from stale identity')
                return None

            delta = time.monotonic() - self.last_new_identity_time
            if delta < self.COOLDOWN_TIME and self.try_num == 1:
                err = ('Retried with new circuit %d times (max) within last '
                       '%d seconds.' % (self.MAX_TRIES, self.COOLDOWN_TIME))
                print('new_identity:', err)
                return err
            elif delta >= self.COOLDOWN_TIME:
                self.try_num = 1

            try:
                #port = settings.tor_control_port
                port = self.tor_port
                with stem.control.Controller.from_port(port=port) as controller:
                    controller.authenticate('')
                    print('new_identity: Getting new identity')
                    controller.signal(stem.Signal.NEWNYM)
                    print('new_identity: NEWNYM signal sent')
                    self.last_new_identity_time = time.monotonic()
                self.refresh_tor_connection_pool()
            except stem.SocketError:
                traceback.print_exc()
                return 'Failed to connect to Tor control port.'
            finally:
                original_try_num = self.try_num
                self.try_num += 1
                if self.try_num > self.MAX_TRIES:
                    self.try_num = 1

            # If we do the request right after second new identity it won't
            # be a new IP, based on experiments.
            # Not necessary after first new identity
            if original_try_num > 1:
                print('Sleeping for 7 seconds before retrying request')
                time.sleep(7)   # experimentally determined minimum

            return None
        finally:
            self.new_identity_lock.release()


tor_manager = TorManager()


def get_pool(use_tor):
    if not use_tor:
        return connection_pool
    return tor_manager.get_tor_connection_pool()


class HTTPAsymmetricCookieProcessor(urllib.request.BaseHandler):
    '''Separate cookiejars for receiving and sending'''
    def __init__(self, cookiejar_send=None, cookiejar_receive=None):
        import http.cookiejar
        self.cookiejar_send = cookiejar_send
        self.cookiejar_receive = cookiejar_receive

    def http_request(self, request):
        if self.cookiejar_send is not None:
            self.cookiejar_send.add_cookie_header(request)
        return request

    def http_response(self, request, response):
        if self.cookiejar_receive is not None:
            self.cookiejar_receive.extract_cookies(response, request)
        return response

    https_request = http_request
    https_response = http_response


class FetchError(Exception):
    def __init__(self, code, reason='', ip=None, error_message=None):
        if error_message:
            string = code + ' ' + reason + ': ' + error_message
        else:
            string = 'HTTP error during request: ' + code + ' ' + reason
        Exception.__init__(self, string)
        self.code = code
        self.reason = reason
        self.ip = ip
        self.error_message = error_message


def decode_content(content, encoding_header):
    encodings = encoding_header.replace(' ', '').split(',')
    for encoding in reversed(encodings):
        if encoding == 'identity':
            continue
        if encoding == 'br':
            content = brotli.decompress(content)
        elif encoding == 'gzip':
            content = gzip.decompress(content)
    return content


def fetch_url_response(url, headers=(), timeout=15, data=None,
                       cookiejar_send=None, cookiejar_receive=None,
                       use_tor=True, max_redirects=None):
    '''
    returns response, cleanup_function
    When cookiejar_send is set to a CookieJar object,
     those cookies will be sent in the request (but cookies in response will not be merged into it)
    When cookiejar_receive is set to a CookieJar object,
     cookies received in the response will be merged into the object (nothing will be sent from it)
    When both are set to the same object, cookies will be sent from the object,
     and response cookies will be merged into it.
    '''
    headers = dict(headers)     # Note: Calling dict() on a dict will make a copy
    if have_brotli:
        headers['Accept-Encoding'] = 'gzip, br'
    else:
        headers['Accept-Encoding'] = 'gzip'

    # prevent python version being leaked by urllib if User-Agent isn't provided
    #  (urllib will use ex. Python-urllib/3.6 otherwise)
    if 'User-Agent' not in headers and 'user-agent' not in headers and 'User-agent' not in headers:
        headers['User-Agent'] = 'Python-urllib'

    method = "GET"
    if data is not None:
        method = "POST"
        if isinstance(data, str):
            data = data.encode('utf-8')
        elif not isinstance(data, bytes):
            data = urllib.parse.urlencode(data).encode('utf-8')


    if cookiejar_send is not None or cookiejar_receive is not None:     # Use urllib
        req = urllib.request.Request(url, data=data, headers=headers)

        cookie_processor = HTTPAsymmetricCookieProcessor(cookiejar_send=cookiejar_send, cookiejar_receive=cookiejar_receive)

        #if use_tor and settings.route_tor:
        if use_tor:
            opener = urllib.request.build_opener(sockshandler.SocksiPyHandler(socks.PROXY_TYPE_SOCKS5, "127.0.0.1", self.tor_port), cookie_processor)
        else:
            opener = urllib.request.build_opener(cookie_processor)

        response = opener.open(req, timeout=timeout)
        cleanup_func = (lambda r: None)

    else:           # Use a urllib3 pool. Cookies can't be used since urllib3 doesn't have easy support for them.
        # default: Retry.DEFAULT = Retry(3)
        # (in connectionpool.py in urllib3)
        # According to the documentation for urlopen, a redirect counts as a
        # retry. So there are 3 redirects max by default.
        if max_redirects:
            retries = urllib3.Retry(3+max_redirects, redirect=max_redirects, raise_on_redirect=False)
        else:
            retries = urllib3.Retry(3, raise_on_redirect=False)
        #pool = get_pool(use_tor and settings.route_tor)
        pool = get_pool(use_tor)
        try:
            response = pool.request(method, url, headers=headers, body=data,
                                    timeout=timeout, preload_content=False,
                                    decode_content=False, retries=retries)
            response.retries = retries
        except urllib3.exceptions.MaxRetryError as e:
            exception_cause = e.__context__.__context__
            # if (isinstance(exception_cause, socks.ProxyConnectionError)
            #         and settings.route_tor):
            if isinstance(exception_cause, socks.ProxyConnectionError):
                msg = ('Failed to connect to Tor. Check that Tor is open and '
                       'that your internet connection is working.\n\n'
                       + str(e))
                raise FetchError('502', reason='Bad Gateway',
                                 error_message=msg)
            elif isinstance(e.__context__,
                            urllib3.exceptions.NewConnectionError):
                msg = 'Failed to establish a connection.\n\n' + str(e)
                raise FetchError(
                    '502', reason='Bad Gateway',
                     error_message=msg)
            else:
                raise
        cleanup_func = (lambda r: r.release_conn())

    return response, cleanup_func


def fetch_url(url, headers=(), timeout=15, report_text=None, data=None,
              cookiejar_send=None, cookiejar_receive=None, use_tor=True,
              debug_name=None):
    while True:
        start_time = time.monotonic()

        response, cleanup_func = fetch_url_response(
            url, headers, timeout=timeout, data=data,
            cookiejar_send=cookiejar_send, cookiejar_receive=cookiejar_receive,
            use_tor=use_tor)
        response_time = time.monotonic()

        content = response.read()

        read_finish = time.monotonic()

        cleanup_func(response)  # release_connection for urllib3
        content = decode_content(
            content,
            response.getheader('Content-Encoding', default='identity'))


        if response.status == 429 or (
            response.status == 302 and (response.getheader('Location') == url
                or response.getheader('Location').startswith(
                       'https://www.google.com/sorry/index'
                   )
            )
        ):
            print(response.status, response.reason, response.getheaders())
            ip = re.search(
                br'IP address: ((?:[\da-f]*:)+[\da-f]+|(?:\d+\.)+\d+)',
                content)
            ip = ip.group(1).decode('ascii') if ip else None
            if not ip:
                ip = re.search(r'IP=((?:\d+\.)+\d+)',
                               response.getheader('Set-Cookie') or '')
                ip = ip.group(1) if ip else None

            # don't get new identity if we're not using Tor
            if not use_tor:
                raise FetchError('429', reason=response.reason, ip=ip)

            print('Error: YouTube blocked the request because the Tor exit node is overutilized. Exit node IP address: %s' % ip)

            # get new identity
            error = tor_manager.new_identity(start_time)
            if error:
                raise FetchError(
                    '429', reason=response.reason, ip=ip,
                    error_message='Automatic circuit change: ' + error)
            else:
                continue # retry now that we have new identity

        elif response.status >= 400:
            raise FetchError(str(response.status), reason=response.reason,
                             ip=None)
        break

    if report_text:
        print(report_text, '    Latency:', round(response_time - start_time, 3), '    Read time:', round(read_finish - response_time,3))

    return content


def head(url, use_tor=False, report_text=None, max_redirects=10):
    #pool = get_pool(use_tor and settings.route_tor)
    pool = get_pool(use_tor)
    start_time = time.monotonic()

    # default: Retry.DEFAULT = Retry(3)
    # (in connectionpool.py in urllib3)
    # According to the documentation for urlopen, a redirect counts as a retry
    # So there are 3 redirects max by default. Let's change that
    # to 10 since googlevideo redirects a lot.
    retries = urllib3.Retry(
        3+max_redirects,
        redirect=max_redirects,
        raise_on_redirect=False)
    headers = {'User-Agent': 'Python-urllib'}
    response = pool.request('HEAD', url, headers=headers, retries=retries)
    if report_text:
        print(
            report_text,
            '    Latency:',
            round(time.monotonic() - start_time, 3))
    return response


mobile_user_agent = 'Mozilla/5.0 (Linux; Android 7.0; Redmi Note 4 Build/NRD90M) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/69.0.3497.100 Mobile Safari/537.36'
mobile_ua = (('User-Agent', mobile_user_agent),)
desktop_user_agent = 'Mozilla/5.0 (Windows NT 6.1; rv:52.0) Gecko/20100101 Firefox/52.0'
desktop_ua = (('User-Agent', desktop_user_agent),)
json_header = (('Content-Type', 'application/json'),)
desktop_xhr_headers = (
    ('Accept', '*/*'),
    ('Accept-Language', 'en-US,en;q=0.5'),
    ('X-YouTube-Client-Name', '1'),
    ('X-YouTube-Client-Version', '2.20180830'),
) + desktop_ua
mobile_xhr_headers = (
    ('Accept', '*/*'),
    ('Accept-Language', 'en-US,en;q=0.5'),
    ('X-YouTube-Client-Name', '2'),
    ('X-YouTube-Client-Version', '2.20180830'),
) + mobile_ua


class RateLimitedQueue(gevent.queue.Queue):
    ''' Does initial_burst (def. 30) at first, then alternates between waiting waiting_period (def. 5) seconds and doing subsequent_bursts (def. 10) queries. After 5 seconds with nothing left in the queue, resets rate limiting. '''

    def __init__(self, initial_burst=30, waiting_period=5, subsequent_bursts=10):
        self.initial_burst = initial_burst
        self.waiting_period = waiting_period
        self.subsequent_bursts = subsequent_bursts

        self.count_since_last_wait = 0
        self.surpassed_initial = False

        self.lock = gevent.lock.BoundedSemaphore(1)
        self.currently_empty = False
        self.empty_start = 0
        gevent.queue.Queue.__init__(self)

    def get(self):
        self.lock.acquire()     # blocks if another greenlet currently has the lock
        if self.count_since_last_wait >= self.subsequent_bursts and self.surpassed_initial:
            gevent.sleep(self.waiting_period)
            self.count_since_last_wait = 0

        elif self.count_since_last_wait >= self.initial_burst and not self.surpassed_initial:
            self.surpassed_initial = True
            gevent.sleep(self.waiting_period)
            self.count_since_last_wait = 0

        self.count_since_last_wait += 1

        if not self.currently_empty and self.empty():
            self.currently_empty = True
            self.empty_start = time.monotonic()

        item = gevent.queue.Queue.get(self)     # blocks when nothing left

        if self.currently_empty:
            if time.monotonic() - self.empty_start >= self.waiting_period:
                self.count_since_last_wait = 0
                self.surpassed_initial = False

            self.currently_empty = False

        self.lock.release()

        return item


def download_thumbnail(save_directory, video_id):
    url = "https://i.ytimg.com/vi/" + video_id + "/mqdefault.jpg"
    save_location = os.path.join(save_directory, video_id + ".jpg")
    try:
        thumbnail = fetch_url(url, report_text="Saved thumbnail: " + video_id)
    except urllib.error.HTTPError as e:
        print("Failed to download thumbnail for " + video_id + ": " + str(e))
        return False
    try:
        f = open(save_location, 'wb')
    except FileNotFoundError:
        os.makedirs(save_directory, exist_ok=True)
        f = open(save_location, 'wb')
    f.write(thumbnail)
    f.close()
    return True


def download_thumbnails(save_directory, ids):
    if not isinstance(ids, (list, tuple)):
        ids = list(ids)
    # only do 5 at a time
    # do the n where n is divisible by 5
    i = -1
    for i in range(0, int(len(ids)/5) - 1 ):
        gevent.joinall([gevent.spawn(download_thumbnail, save_directory, ids[j]) for j in range(i*5, i*5 + 5)])
    # do the remainders (< 5)
    gevent.joinall([gevent.spawn(download_thumbnail, save_directory, ids[j]) for j in range(i*5 + 5, len(ids))])


def dict_add(*dicts):
    for dictionary in dicts[1:]:
        dicts[0].update(dictionary)
    return dicts[0]


def video_id(url):
    url_parts = urllib.parse.urlparse(url)
    return urllib.parse.parse_qs(url_parts.query)['v'][0]


# default, sddefault, mqdefault, hqdefault, hq720
def get_thumbnail_url(video_id):
    #return settings.img_prefix + "https://i.ytimg.com/vi/" + video_id + "/mqdefault.jpg"
    return "https://i.ytimg.com/vi/" + video_id + "/mqdefault.jpg"


def seconds_to_timestamp(seconds):
    seconds = int(seconds)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    if hours != 0:
        timestamp = str(hours) + ":"
        timestamp += str(minutes).zfill(2)  # zfill pads with zeros
    else:
        timestamp = str(minutes)

    timestamp += ":" + str(seconds).zfill(2)
    return timestamp


def update_query_string(query_string, items):
    parameters = urllib.parse.parse_qs(query_string)
    parameters.update(items)
    return urllib.parse.urlencode(parameters, doseq=True)


YOUTUBE_DOMAINS = ('youtube.com', 'youtu.be', 'youtube-nocookie.com')
YOUTUBE_URL_RE_STR = r'https?://(?:[a-zA-Z0-9_-]*\.)?(?:'
YOUTUBE_URL_RE_STR += r'|'.join(map(re.escape, YOUTUBE_DOMAINS))
YOUTUBE_URL_RE_STR += r')(?:/[^"]*)?'
YOUTUBE_URL_RE = re.compile(YOUTUBE_URL_RE_STR)


def prefix_url(url):
    if url is None:
        return None
    url = url.lstrip('/')     # some urls have // before them, which has a special meaning
    return '/' + url


def left_remove(string, substring):
    '''removes substring from the start of string, if present'''
    if string.startswith(substring):
        return string[len(substring):]
    return string


def concat_or_none(*strings):
    '''Concatenates strings. Returns None if any of the arguments are None'''
    result = ''
    for string in strings:
        if string is None:
            return None
        result += string
    return result


def prefix_urls(item):
    # if settings.proxy_images:
    #     try:
    #         item['thumbnail'] = prefix_url(item['thumbnail'])
    #     except KeyError:
    #         pass

    try:
        item['author_url'] = prefix_url(item['author_url'])
    except KeyError:
        pass


def add_extra_html_info(item):
    if item['type'] == 'video':
        item['url'] = (URL_ORIGIN + '/watch?v=' + item['id']) if item.get('id') else None

        video_info = {}
        for key in ('id', 'title', 'author', 'duration', 'author_id'):
            try:
                video_info[key] = item[key]
            except KeyError:
                video_info[key] = None

        item['video_info'] = json.dumps(video_info)

    elif item['type'] == 'playlist' and item['playlist_type'] == 'radio':
        item['url'] = concat_or_none(
            URL_ORIGIN,
            '/watch?v=', item['first_video_id'],
            '&list=', item['id']
        )
    elif item['type'] == 'playlist':
        item['url'] = concat_or_none(URL_ORIGIN, '/playlist?list=', item['id'])
    elif item['type'] == 'channel':
        item['url'] = concat_or_none(URL_ORIGIN, "/channel/", item['id'])

    if item.get('author_id') and 'author_url' not in item:
        item['author_url'] = URL_ORIGIN + '/channel/' + item['author_id']


def check_gevent_exceptions(*tasks):
    for task in tasks:
        if task.exception:
            raise task.exception


# https://stackoverflow.com/a/62888
replacement_map = collections.OrderedDict([
    ('<', '_'),
    ('>', '_'),
    (': ', ' - '),
    (':', '-'),
    ('"', "'"),
    ('/', '_'),
    ('\\', '_'),
    ('|', '-'),
    ('?', ''),
    ('*', '_'),
    ('\t', ' '),
])

DOS_names = {'con', 'prn', 'aux', 'nul', 'com0', 'com1', 'com2', 'com3',
             'com4', 'com5', 'com6', 'com7', 'com8', 'com9', 'lpt0',
             'lpt1', 'lpt2', 'lpt3', 'lpt4', 'lpt5', 'lpt6', 'lpt7',
             'lpt8', 'lpt9'}


def to_valid_filename(name):
    '''Changes the name so it's valid on Windows, Linux, and Mac'''
    # See https://docs.microsoft.com/en-us/windows/win32/fileio/naming-a-file
    # for Windows specs

    # Additional recommendations for Linux:
    # https://dwheeler.com/essays/fixing-unix-linux-filenames.html#standards

    # remove control characters
    name = re.sub(r'[\x00-\x1f]', '_', name)

    # reserved characters
    for reserved_char, replacement in replacement_map.items():
        name = name.replace(reserved_char, replacement)

    # check for all periods/spaces
    if all(c == '.' or c == ' ' for c in name):
        name = '_'*len(name)

    # remove trailing periods and spaces
    name = name.rstrip('. ')

    # check for reserved DOS names, such as nul or nul.txt
    base_ext_parts = name.rsplit('.', maxsplit=1)
    if base_ext_parts[0].lower() in DOS_names:
        base_ext_parts[0] += '_'
    name = '.'.join(base_ext_parts)

    # check for blank name
    if name == '':
        name = '_'

    # check if name begins with a hyphen, period, or space
    if name[0] in ('-', '.', ' '):
        name = '_' + name

    return name


def strip_non_ascii(string):
    ''' Returns the string without non ASCII characters'''
    stripped = (c for c in string if 0 < ord(c) < 127)
    return ''.join(stripped)


def time_utc_isoformat(string):
    t = datetime.strptime(string, '%Y-%m-%d')
    t = t.astimezone().isoformat()
    return t


"""
    Video abuse report reasons resource implementation.
"""
from typing import Optional, Union

from pyyoutube.resources.base_resource import Resource
from pyyoutube.models import VideoAbuseReportReasonListResponse
from pyyoutube.utils.params_checker import enf_parts


class VideoAbuseReportReasonsResource(Resource):
    """A videoAbuseReportReason resource contains information about a reason that a video would be flagged
    for containing abusive content.

    References: https://developers.google.com/youtube/v3/docs/videoAbuseReportReasons
    """

    def list(
        self,
        parts: Optional[Union[str, list, tuple, set]] = None,
        hl: Optional[str] = None,
        return_json: bool = False,
        **kwargs: Optional[dict],
    ) -> Union[dict, VideoAbuseReportReasonListResponse]:
        """Retrieve a list of reasons that can be used to report abusive videos.

        Args:
            parts:
                Comma-separated list of one or more channel resource properties.
                Accepted values: id,snippet
            hl:
                Specifies the language that should be used for text values in the API response.
                The default value is en_US.
            return_json:
                Type for returned data. If you set True JSON data will be returned.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            reasons data.
        """
        params = {
            "part": enf_parts(resource="videoAbuseReportReasons", value=parts),
            "hl": hl,
            **kwargs,
        }
        response = self._client.request(path="videoAbuseReportReasons", params=params)
        data = self._client.parse_response(response=response)
        return (
            data if return_json else VideoAbuseReportReasonListResponse.from_dict(data)
        )



"""
    Search resource implementation.
"""
from typing import Optional, Union

from pyyoutube.resources.base_resource import Resource
from pyyoutube.models import SearchListResponse
from pyyoutube.utils.params_checker import enf_parts


class SearchResource(Resource):
    """A search result contains information about a YouTube video, channel, or playlist
    that matches the search parameters specified in an API request

    References: https://developers.google.com/youtube/v3/docs/search
    """

    def list(
        self,
        parts: Optional[Union[str, list, tuple, set]] = None,
        for_content_owner: Optional[bool] = None,
        for_developer: Optional[bool] = None,
        for_mine: Optional[bool] = None,
        related_to_video_id: Optional[str] = None,
        channel_id: Optional[str] = None,
        channel_type: Optional[str] = None,
        event_type: Optional[str] = None,
        location: Optional[str] = None,
        location_radius: Optional[str] = None,
        max_results: Optional[int] = None,
        on_behalf_of_content_owner: Optional[str] = None,
        order: Optional[str] = None,
        page_token: Optional[str] = None,
        published_after: Optional[str] = None,
        published_before: Optional[str] = None,
        q: Optional[str] = None,
        region_code: Optional[str] = None,
        relevance_language: Optional[str] = None,
        safe_search: Optional[str] = None,
        topic_id: Optional[str] = None,
        type: Optional[Union[str, list, tuple, set]] = None,
        video_caption: Optional[str] = None,
        video_category_id: Optional[str] = None,
        video_definition: Optional[str] = None,
        video_dimension: Optional[str] = None,
        video_duration: Optional[str] = None,
        video_embeddable: Optional[str] = None,
        video_license: Optional[str] = None,
        video_paid_product_placement: Optional[str] = None,
        video_syndicated: Optional[str] = None,
        video_type: Optional[str] = None,
        return_json: bool = False,
        **kwargs: Optional[dict],
    ) -> Union[dict, SearchListResponse]:
        """Returns a collection of search results that match the query parameters specified in the API request.

        Notes:
            Search API is very complex. If you want to search, You may need to read the parameter description
            with the docs: https://developers.google.com/youtube/v3/docs/search/list#parameters

        Args:
            parts:
                Comma-separated list of one or more channel resource properties.
                Accepted values: snippet
            for_content_owner:
                Parameter restricts the search to only retrieve videos owned by the content owner
                identified by the onBehalfOfContentOwner parameter.
            for_developer:
                Parameter restricts the search to only retrieve videos uploaded via the developer's
                application or website.
            for_mine:
                Parameter restricts the search to only retrieve videos owned by the authenticated user.
            related_to_video_id:
                Parameter retrieves a list of videos that are related to the video that the parameter value identifies.
                Deprecated at [2023.08.07](https://developers.google.com/youtube/v3/revision_history#august-7,-2023)
            channel_id:
                Indicates that the API response should only contain resources created by the channel.
            channel_type:
                Parameter lets you restrict a search to a particular type of channel.
                Acceptable values are:
                    - any: Return all channels.
                    - show: Only retrieve shows.
            event_type:
                Parameter restricts a search to broadcast events.
                Acceptable values are:
                    - completed: Only include completed broadcasts.
                    - live: Only include active broadcasts.
                    - upcoming: Only include upcoming broadcasts.
            location:
                Parameter value identifies the point at the center of the area.
            location_radius:
                Specifies the maximum distance that the location associated with a video can be from
                that point for the video to still be included in the search results.
            max_results:
                The parameter specifies the maximum number of items that should be returned
                the result set.
                Acceptable values are 0 to 50, inclusive. The default value is 5.
            on_behalf_of_content_owner:
                The onBehalfOfContentOwner parameter indicates that the request's authorization
                credentials identify a YouTube CMS user who is acting on behalf of the content
                owner specified in the parameter value. This parameter is intended for YouTube
                content partners that own and manage many difference YouTube channels. It allows
                content owners to authenticate once and get access to all their video and channel
                data, without having to provide authentication credentials for each individual channel.
                The CMS account that the user authenticates with must be linked to the specified YouTube content owner.
            order:
                Specifies the method that will be used to order resources in the API response.
                The default value is relevance.
                Acceptable values are:
                    - date: Resources are sorted in reverse chronological order based on the date they were created.
                    - rating: Resources are sorted from highest to lowest rating.
                    - relevance: Resources are sorted based on their relevance to the search query.
                    - title: Resources are sorted alphabetically by title.
                    - videoCount: Channels are sorted in descending order of their number of uploaded videos.
                    - viewCount: Resources are sorted from highest to lowest number of views.
                    For live broadcasts, videos are sorted by number of concurrent viewers while the broadcasts
                    are ongoing.
            page_token:
                The parameter identifies a specific page in the result set that should be returned.
            published_after:
                Indicates that the API response should only contain resources created at or after the specified time.
            published_before:
                Indicates that the API response should only contain resources created before or at the specified time.
            q:
                Specifies the query term to search for.
            region_code:
                Instructs the API to return search results for videos that can be viewed in the specified country.
            relevance_language:
                Instructs the API to return search results that are most relevant to the specified language.
            safe_search:
                Indicates whether the search results should include restricted content as well as standard content.
                Acceptable values are:
                    - moderate: YouTube will filter some content from search results and, at the least,
                        will filter content that is restricted in your locale. Based on their content, search
                        results could be removed from search results or demoted in search results.
                        This is the default parameter value.
                    - none: YouTube will not filter the search result set.
                    - strict: YouTube will try to exclude all restricted content from the search result set.
                        Based on their content, search results could be removed from search results or
                        demoted in search results.
            topic_id:
                Indicates that the API response should only contain resources associated with the specified topic.
            type:
                Parameter restricts a search query to only retrieve a particular type of resource.
                The value is a comma-separated list of resource types.
                Acceptable values are: channel,playlist,video
            video_caption:
                Indicates whether the API should filter video search results based on whether they have captions.
                Acceptable values are:
                    - any: Do not filter results based on caption availability.
                    - closedCaption: Only include videos that have captions.
                    - none: Only include videos that do not have captions.
            video_category_id:
                Parameter filters video search results based on their category.
            video_definition:
                Parameter lets you restrict a search to only include either high definition (HD) or
                standard definition (SD) videos.
                Acceptable values are:
                    - any: Return all videos, regardless of their resolution.
                    - high: Only retrieve HD videos.
                    - standard: Only retrieve videos in standard definition.
            video_dimension:
                Parameter lets you restrict a search to only retrieve 2D or 3D videos.
                Acceptable values are:
                    - 2d: Restrict search results to exclude 3D videos.
                    - 3d: Restrict search results to only include 3D videos.
                    - any: Include both 3D and non-3D videos in returned results. This is the default value.
            video_duration:
                Parameter filters video search results based on their duration.
                Acceptable values are:
                    - any: Do not filter video search results based on their duration. This is the default value.
                    - long: Only include videos longer than 20 minutes.
                    - medium: Only include videos that are between four and 20 minutes long (inclusive).
                    - short: Only include videos that are less than four minutes long.
            video_embeddable:
                Parameter lets you to restrict a search to only videos that can be embedded into a webpage.
                Acceptable values are:
                    - any: Return all videos, embeddable or not.
                    - true: Only retrieve embeddable videos.
            video_license:
                Parameter filters search results to only include videos with a particular license.
                Acceptable values are:
                    - any – Return all videos, regardless of which license they have, that match the query parameters.
                    - creativeCommon – Only return videos that have a Creative Commons license.
                        Users can reuse videos with this license in other videos that they create. Learn more.
                    - youtube – Only return videos that have the standard YouTube license.
            video_paid_product_placement:
                Parameter filters search results to only include videos that the creator has denoted as
                having a paid promotion.
                Acceptable values are:
                    - any – Return all videos, regardless of whether they contain paid promotions.
                    - true – Only retrieve videos with paid promotions.
            video_syndicated:
                Parameter lets you to restrict a search to only videos that can be played outside youtube.com.
                Acceptable values are:
                    - any: Return all videos, syndicated or not.
                    - true: Only retrieve syndicated videos.
            video_type:
                Parameter lets you restrict a search to a particular type of videos.
                Acceptable values are:
                    - any: Return all videos.
                    - episode: Only retrieve episodes of shows.
                    - movie: Only retrieve movies.
            return_json:
                Type for returned data. If you set True JSON data will be returned.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.
        Returns:
            Search result data

        """

        params = {
            "part": enf_parts(resource="search", value=parts),
            "channelId": channel_id,
            "channelType": channel_type,
            "eventType": event_type,
            "location": location,
            "locationRadius": location_radius,
            "maxResults": max_results,
            "onBehalfOfContentOwner": on_behalf_of_content_owner,
            "order": order,
            "pageToken": page_token,
            "publishedAfter": published_after,
            "publishedBefore": published_before,
            "q": q,
            "regionCode": region_code,
            "relevanceLanguage": relevance_language,
            "safeSearch": safe_search,
            "topicId": topic_id,
            "type": type,
            "videoCaption": video_caption,
            "videoCategoryId": video_category_id,
            "videoDefinition": video_definition,
            "videoDimension": video_dimension,
            "videoDuration": video_duration,
            "videoEmbeddable": video_embeddable,
            "videoLicense": video_license,
            "videoPaidProductPlacement": video_paid_product_placement,
            "videoSyndicated": video_syndicated,
            "videoType": video_type,
            **kwargs,
        }

        if for_content_owner is not None:
            params["forContentOwner"] = for_content_owner
        elif for_developer is not None:
            params["forDeveloper"] = for_developer
        elif for_mine is not None:
            params["forMine"] = for_mine
        elif related_to_video_id is not None:
            params["relatedToVideoId"] = related_to_video_id

        response = self._client.request(path="search", params=params)
        data = self._client.parse_response(response=response)
        return data if return_json else SearchListResponse.from_dict(data)



"""
    Playlist resource implementation.
"""

from typing import Optional, Union

from pyyoutube.error import PyYouTubeException, ErrorCode, ErrorMessage
from pyyoutube.resources.base_resource import Resource
from pyyoutube.models import Playlist, PlaylistListResponse
from pyyoutube.utils.params_checker import enf_comma_separated, enf_parts


class PlaylistsResource(Resource):
    """A playlist resource represents a YouTube playlist.

    References: https://developers.google.com/youtube/v3/docs/playlists
    """

    def list(
        self,
        parts: Optional[Union[str, list, tuple, set]] = None,
        channel_id: Optional[str] = None,
        playlist_id: Optional[Union[str, list, tuple, set]] = None,
        mine: Optional[bool] = None,
        hl: Optional[str] = None,
        max_results: Optional[int] = None,
        on_behalf_of_content_owner: Optional[str] = None,
        on_behalf_of_content_owner_channel: Optional[str] = None,
        page_token: Optional[str] = None,
        return_json: bool = False,
        **kwargs: Optional[dict],
    ) -> Union[dict, PlaylistListResponse]:
        """Returns a collection of playlists that match the API request parameters.

        Args:
            parts:
                Comma-separated list of one or more channel resource properties.
                Accepted values: id,contentDetails,localizations,player,snippet,status
            channel_id:
                Indicates that the API should only return the specified channel's playlists.
            playlist_id:
                Specifies a comma-separated list of the YouTube playlist ID(s) for the resource(s)
                that are being retrieved.
            mine:
                Set this parameter's value to true to instruct the API to only return playlists
                owned by the authenticated user.
            hl:
                The hl parameter instructs the API to retrieve localized resource metadata for
                a specific application language that the YouTube website supports.
                The parameter value must be a language code included in the list returned by the
                i18nLanguages.list method.
            max_results:
                The parameter specifies the maximum number of items that should be returned
                the result set.
                Acceptable values are 0 to 50, inclusive. The default value is 5.
            on_behalf_of_content_owner:
                The onBehalfOfContentOwner parameter indicates that the request's authorization
                credentials identify a YouTube CMS user who is acting on behalf of the content
                owner specified in the parameter value. This parameter is intended for YouTube
                content partners that own and manage many difference YouTube channels. It allows
                content owners to authenticate once and get access to all their video and channel
                data, without having to provide authentication credentials for each individual channel.
                The CMS account that the user authenticates with must be linked to the specified YouTube content owner.
            on_behalf_of_content_owner_channel:
                The onBehalfOfContentOwnerChannel parameter specifies the YouTube channel ID of the channel
                to which a video is being added. This parameter is required when a request specifies a value
                for the onBehalfOfContentOwner parameter, and it can only be used in conjunction with that
                parameter. In addition, the request must be authorized using a CMS account that is linked to
                the content owner that the onBehalfOfContentOwner parameter specifies. Finally, the channel
                that the onBehalfOfContentOwnerChannel parameter value specifies must be linked to the content
                owner that the onBehalfOfContentOwner parameter specifies.
            page_token:
                The parameter identifies a specific page in the result set that should be returned.
            return_json:
                Type for returned data. If you set True JSON data will be returned.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Playlist data.
        Raises:
            PyYouTubeException: Missing filter parameter.
        """

        params = {
            "part": enf_parts(resource="playlists", value=parts),
            "hl": hl,
            "maxResults": max_results,
            "onBehalfOfContentOwner": on_behalf_of_content_owner,
            "onBehalfOfContentOwnerChannel": on_behalf_of_content_owner_channel,
            "pageToken": page_token,
            **kwargs,
        }
        if channel_id is not None:
            params["channelId"] = channel_id
        elif playlist_id is not None:
            params["id"] = enf_comma_separated(field="playlist_id", value=playlist_id)
        elif mine is not None:
            params["mine"] = mine
        else:
            raise PyYouTubeException(
                ErrorMessage(
                    status_code=ErrorCode.MISSING_PARAMS,
                    message=f"Specify at least one of channel_id, playlist_id or mine",
                )
            )
        response = self._client.request(path="playlists", params=params)
        data = self._client.parse_response(response=response)
        return data if return_json else PlaylistListResponse.from_dict(data)

    def insert(
        self,
        body: Union[dict, Playlist],
        parts: Optional[Union[str, list, tuple, set]] = None,
        on_behalf_of_content_owner: Optional[str] = None,
        on_behalf_of_content_owner_channel: Optional[str] = None,
        return_json: bool = False,
        **kwargs: Optional[dict],
    ) -> Union[dict, Playlist]:
        """Creates a playlist.

        Args:
            body:
                Provide playlist data in the request body. You can give dataclass or just a dict with data.
            parts:
                The part parameter serves two purposes in this operation. It identifies the properties
                that the write operation will set as well as the properties that the API response will include.
            on_behalf_of_content_owner:
                The onBehalfOfContentOwner parameter indicates that the request's authorization
                credentials identify a YouTube CMS user who is acting on behalf of the content
                owner specified in the parameter value. This parameter is intended for YouTube
                content partners that own and manage many difference YouTube channels. It allows
                content owners to authenticate once and get access to all their video and channel
                data, without having to provide authentication credentials for each individual channel.
                The CMS account that the user authenticates with must be linked to the specified YouTube content owner.
            on_behalf_of_content_owner_channel:
                The onBehalfOfContentOwnerChannel parameter specifies the YouTube channel ID of the channel
                to which a video is being added. This parameter is required when a request specifies a value
                for the onBehalfOfContentOwner parameter, and it can only be used in conjunction with that
                parameter. In addition, the request must be authorized using a CMS account that is linked to
                the content owner that the onBehalfOfContentOwner parameter specifies. Finally, the channel
                that the onBehalfOfContentOwnerChannel parameter value specifies must be linked to the content
                owner that the onBehalfOfContentOwner parameter specifies.
            return_json:
                Type for returned data. If you set True JSON data will be returned.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.
        Returns:
            playlist data.
        """

        params = {
            "part": enf_parts(resource="playlists", value=parts),
            "onBehalfOfContentOwner": on_behalf_of_content_owner,
            "onBehalfOfContentOwnerChannel": on_behalf_of_content_owner_channel,
            **kwargs,
        }
        response = self._client.request(
            method="POST", path="playlists", params=params, json=body
        )
        data = self._client.parse_response(response=response)
        return data if return_json else Playlist.from_dict(data)

    def update(
        self,
        body: Union[dict, Playlist],
        parts: Optional[Union[str, list, tuple, set]] = None,
        on_behalf_of_content_owner: Optional[str] = None,
        return_json: bool = False,
        **kwargs: Optional[dict],
    ) -> Union[dict, Playlist]:
        """Modifies a playlist.

        Args:
            body:
                Provide playlist data in the request body. You can give dataclass or just a dict with data.
            parts:
                The part parameter serves two purposes in this operation. It identifies the properties
                that the write operation will set as well as the properties that the API response will include.
            on_behalf_of_content_owner:
                The onBehalfOfContentOwner parameter indicates that the request's authorization
                credentials identify a YouTube CMS user who is acting on behalf of the content
                owner specified in the parameter value. This parameter is intended for YouTube
                content partners that own and manage many difference YouTube channels. It allows
                content owners to authenticate once and get access to all their video and channel
                data, without having to provide authentication credentials for each individual channel.
                The CMS account that the user authenticates with must be linked to the specified YouTube content owner.
            return_json:
                Type for returned data. If you set True JSON data will be returned.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Playlist updated data.

        """
        params = {
            "part": enf_parts(resource="playlists", value=parts),
            "onBehalfOfContentOwner": on_behalf_of_content_owner,
            **kwargs,
        }
        response = self._client.request(
            method="PUT", path="playlists", params=params, json=body
        )
        data = self._client.parse_response(response=response)
        return data if return_json else Playlist.from_dict(data)

    def delete(
        self,
        playlist_id: str,
        on_behalf_of_content_owner: Optional[str] = None,
        **kwargs: Optional[dict],
    ) -> bool:
        """Deletes a playlist.

        Args:
            playlist_id:
                Specifies the YouTube playlist ID for the playlist that is being deleted.
            on_behalf_of_content_owner:
                The onBehalfOfContentOwner parameter indicates that the request's authorization
                credentials identify a YouTube CMS user who is acting on behalf of the content
                owner specified in the parameter value. This parameter is intended for YouTube
                content partners that own and manage many difference YouTube channels. It allows
                content owners to authenticate once and get access to all their video and channel
                data, without having to provide authentication credentials for each individual channel.
                The CMS account that the user authenticates with must be linked to the specified YouTube content owner.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            playlist delete status

        """
        params = {
            "id": playlist_id,
            "onBehalfOfContentOwner": on_behalf_of_content_owner,
            **kwargs,
        }
        response = self._client.request(
            method="DELETE",
            path="playlists",
            params=params,
        )
        if response.ok:
            return True
        self._client.parse_response(response=response)
